"""Tests the clyngor' Propagator class and the underlying API"""

import textwrap
import clyngor
from .definitions import onlyif_python_support, skipif_no_clingo_module


ASP_CODE = """
p(1).
p(X):- p(X-1) ; X<3.
q(a,b,p(1)).
"""
PYCONSTRAINT_CODE_ARGUMENT = """
#script(python)
from clyngor import Constraint, Variable as V, Main

# Build the constraint on atom b
def formula(inputs) -> bool:
    return inputs['b', ({value_to_avoid},)]

constraint = Constraint(formula, {{('b', (V,))}})

# regular main function
main = Main(propagators=constraint)

#end.

1{{b(1..3)}}1.
"""
PYCONSTRAINT_CODE = PYCONSTRAINT_CODE_ARGUMENT.format(value_to_avoid=2)


class MyPropagator(clyngor.Propagator):

    def __init__(self):
        super().__init__(follow=(('p', [...]), ('q', [...]*3)))
        self.found_atoms = {}

    def on_all_input(self, values:dict):
        for (pred, args), holds in values.items():
            if holds:
                self.found_atoms.setdefault(pred, set()).add(args)


@skipif_no_clingo_module
@onlyif_python_support
def test_the_subclass():
    prop = MyPropagator()
    ctl = prop.run_with(inline=ASP_CODE)
    for answer in ctl:
        pass
    assert prop.found_atoms == {'p': {(1,), (2,)}, 'q': {('a', 'b', ('p', (1,)))}}


@onlyif_python_support
def test_pyconstraint_from_embedded_code():
    possible_values = {1, 2, 3}
    for value_to_avoid in possible_values:
        source = PYCONSTRAINT_CODE_ARGUMENT.format(value_to_avoid=value_to_avoid)
        models = set(clyngor.solve(inline=source, use_clingo_module=False, error_on_warning=True))
        print(source)
        assert len(models) == len(possible_values) - 1
        assert models == {
            frozenset({('b', (val,))})
            for val in possible_values if val != value_to_avoid
        }


@onlyif_python_support
def test_pyconstraint_recursion_problem():
    """Show that the python constraint handles recursive propagations and
    is able to discard models appropriatly.

    if the solving hangs, then probably the constraint implementation
    is blocked by an infinite loop of propagations during discarding.

    if the solving yields inapropriate models, it's (among other)
    because the clauses added by the constraint object are kept
    (which is forced by design by clingo's API).

    Current state : propagation loop is prevented by a state mecanism.
    However, the inapropriate model solving is due to many problems,
    one being clingo's API itself.

    Some ressources for future solving of that feature-killer issue:

    - what we want is not possible. By design.
        https://sourceforge.net/p/potassco/mailman/message/36358095/
    - some other peoples wants it, the solution seems to be external atoms
        https://sourceforge.net/p/potassco/mailman/message/35480361/
    - but external atoms are not handlable during solving through propagator API
        https://potassco.org/clingo/python-api/current/clingo.html#PropagateControl
    - cool find with propagators:
        https://twitter.com/rndmcnlly/status/867605489789489152

    """
    source = textwrap.dedent("""
    #script(python)
    from clyngor import Constraint, Variable as V, Main
    def formula(inputs):  return inputs['p', (2,)]
    constraint = Constraint(formula, {('p', (V,))})
    main = Main(propagators=constraint)
    #end.
    {p(1..3)}.  % powerset of p(X).
    """)
    models = set(clyngor.solve(inline=source, use_clingo_module=False, error_on_warning=True))
    assert models == {
        frozenset(()),
        frozenset({('p', (1,))}),
        frozenset({('p', (2,))}),
        frozenset({('p', (3,))}),
    }, "this may fail because a bug was fixed"
    return
    # the models that are really to be expected:
    #  (and given by a really working implementation)
    assert models == {
        frozenset(()),
        frozenset({('p', (1,))}),
        frozenset({('p', (3,))}),
        frozenset({('p', (1,)), ('p', (3,))}),
    }


@skipif_no_clingo_module
@onlyif_python_support
def test_local_propagator_hidden_by_clingo():
    """Prove that main function (and therefore locally defined propagators)
    is ignored from code if called with clingo module
    """
    clingo_models = set(clyngor.solve(inline=PYCONSTRAINT_CODE, use_clingo_module=True))
    clyngor_models = set(clyngor.solve(inline=PYCONSTRAINT_CODE, use_clingo_module=False))
    assert len(clingo_models) == 3
    assert len(clyngor_models) == 2
    assert clingo_models == {
        frozenset({('b', (1,))}),
        frozenset({('b', (2,))}),
        frozenset({('b', (3,))}),
    }
    assert clyngor_models == {
        frozenset({('b', (1,))}),
        frozenset({('b', (3,))}),
    }


@skipif_no_clingo_module
@onlyif_python_support
def test_pyconstraint_from_python():
    from clyngor import Constraint, Variable as V
    def formula(inputs) -> bool:
        return inputs['b', (2,)]
    constraint = Constraint(formula, {('b', (V,))})
    models = set(constraint.run_with(inline='1{b(1..3)}1.'))
    assert len(models) == 2
    assert models == {
        frozenset({('b', (1,))}),
        frozenset({('b', (3,))}),
    }



"""Various high level definition for client"""


import os
import sys
import math
import tempfile
import functools
import clyngor
from clyngor import parsing

try:
    import clingo
except ImportError:
    clingo = None


class ASPSyntaxError(SyntaxError):
    """This is a SyntaxError, but without the filename at the end of the
    string representation, and with a payload attached.

    Used by solving module to indicate problems in generated files.

    """
    def __init__(self, *args, payload:dict, **kwargs):
        super().__init__(*args, **kwargs)
        self.payload = payload

    def __str__(self):
        return self.msg

class ASPWarning(ValueError):
    """This is a ValueError, with a payload attached to it"""
    def __init__(self, msg:str, payload:dict):
        super().__init__(msg)
        self.payload = payload
        self.atom = payload['atom']


def parse_clingo_output(clingo_output:[str]):
    "Yield answer sets found in given clingo output"
    yield from (answer for anstype, answer
                in parsing.Parser().parse_clasp_output(clingo_output)
                if anstype == 'answer')


def make_hashable(val):
    """Convert lists and sets into tuples and frozensets

    >>> make_hashable(2)
    2
    >>> make_hashable('2')
    '2'
    >>> make_hashable([1, ({2, 3}, [4, 5], {(6, 7): [8, 9]})])
    (1, (frozenset({2, 3}), (4, 5), {(6, 7): (8, 9)}))

    """
    if isinstance(val, (tuple, list)):
        return tuple(map(make_hashable, val))
    elif isinstance(val, (frozenset, set)):
        return frozenset(map(make_hashable, val))
    elif isinstance(val, dict):
        return {make_hashable(k): make_hashable(v) for k, v in val.items()}
    return val


def remove_arguments_quotes(arguments:str):
    """Remove quotes at the beginning and at the end
    of the given arguments in ASP format.

    This function is used by the weak parser, and therefore
    does not need to meet robustness criterion.

    >>> remove_arguments_quotes('a,b')
    'a,b'
    >>> remove_arguments_quotes('"a","b"')
    'a,b'
    >>> remove_arguments_quotes('"a",b')
    'a,b'
    """
    def is_quoted(arg:str) -> bool:
        return (isinstance(arg, str) and len(arg) >= 2
                and arg[0] == '"' and arg[-1] == '"' and arg[-2] != '\\')
    return ','.join(arg[1:-1] if is_quoted(arg) else arg
                    for arg in arguments.split(','))


def clingo_value_to_python(value:object) -> int or str or tuple:
    """Convert a clingo.Symbol object to the python equivalent"""
    if False and str(type(value).__name__) == 'Symbol':
        return clingo_symbol_as_python_value(value)
    elif isinstance(value, int):
        return value
    elif isinstance(value, str):
        return '"' + value + '"'
    elif isinstance(value, (tuple, list)):
        return tuple(map(clingo_value_to_python, value))
    elif type(value).__name__ == 'Symbol':
        try:
            typename = str(value.type).lower()
            assert typename.startswith('symboltype.')
            typename = typename[len('symboltype.'):]
            if typename == 'function':
                if value.arguments:
                    pyvalue = (value.name, tuple(map(clingo_value_to_python, value.arguments)))
                else:
                    pyvalue = value.name
            else:  # for any other type, get its python value directly
                pyvalue = getattr(value, typename)
        except AttributeError as err:  # inf or sup
            if value.type == value.type.Infimum:
                return -math.inf
            elif value.type == value.type.Supremum:
                return math.inf
            else:
                raise err
        if typename == 'string':
            pyvalue = '"' + pyvalue.replace('"', '\\"') + '"'
        return pyvalue
    raise TypeError("Can't handle values like {} of type {}."
                    "".format(value, type(value)))


def clingo_symbol_as_python_value_basefunc(term, typename: str) -> object:
    "Convert a clingo.Symbol object to the python equivalent"
    if typename == 'Function':
        assert term.name is not None
        name = ('-' if term.negative else '') + term.name
        return (name, clingo_value_to_python(term.arguments))
    elif typename == 'String':
        assert term.name is None
        return ('"' + term.string + '"', ())
    elif typename == 'Number':
        assert term.name is None
        return (term.number, ())
    raise TypeError("Can't handle clingo.Symbol like {} of type {}."
                    "".format(term, typename))


def clingo_symbol_as_python_value(term) -> object:
    if clingo:
        types = {
            clingo.SymbolType.Function: 'Function',
            clingo.SymbolType.Number: 'Number',
            clingo.SymbolType.String: 'String',
        }
        return clingo_symbol_as_python_value_basefunc(term, types.get(term.type, term.type))
    else:
        return clingo_symbol_as_python_value_basefunc(term, term.type)

def python_value_to_asp(val:str or int or list or tuple, *, args_of_predicate:bool=False) -> str or tuple:
    """Convert given python value in an ASP format"""
    if isinstance(val, (str, int)):
        return str(val) or '""'
    elif isinstance(val, float):
        return '"' + str(val) + '"'
    elif isinstance(val, tuple):
        assert len(val) == 2, "tuple value should be of size 2: (predicate, args))"
        predicate, args = val
        if args:
            return predicate + '(' + python_value_to_asp(list(args), args_of_predicate=bool(predicate)) + ')'
        else:  # no args
            return predicate
    elif isinstance(val, list):
        ender = ',' if len(val) == 1 and not args_of_predicate else ''
        return ','.join(map(python_value_to_asp, val)) + ender
    raise ValueError("Python value '{}' of type {} is not convertible in ASP.".format(repr(val), type(val)))
# python_value_to_asp.in_predicate = lambda x: python_value_to_asp(x, args_of_predicate=True)


def integers_to_string_atoms(model:iter) -> object:
    """Return an identical structure of (frozen)set, tuple and list, but with integer values as string"""
    if isinstance(model, (list, tuple, frozenset, set)):
        return type(model)(map(integers_to_string_atoms, model))
    elif isinstance(model, dict):
        return {integers_to_string_atoms(k): integers_to_string_atoms(v)
                for k, v in model.items()}
    elif isinstance(model, int):
        return str(model)
    else:
        return model


def answer_set_to_str(answer_set:iter, atom_end:str='', atom_sep:str=' ') -> str:
    """Returns the string representation of given answer set.

    answer_set -- iterable of tuple (predicate, args)
    atom_sep -- string joining the atoms

    """
    return atom_sep.join(generate_answer_set_as_str(answer_set, atom_end=atom_end))


def generate_answer_set_as_str(answer_set:iter, atom_end:str='') -> iter:
    """Yield segment of string describing given answer set.

    answer_set -- iterable of tuple (predicate, args), or dict {predicate: [args]}
    atom_end -- string added to the end of each given atom.

    >>> '.'.join(generate_answer_set_as_str((('a', (1, 2)), ('b', ()))))
    'a(1,2).b'

    """
    if isinstance(answer_set, dict):  # have been created using by_predicate, probably
        answer_set = (
            (predicate, args)
            for predicate, all_args in answer_set.items()
            for args in all_args
        )

    template = '{}({})' + str(atom_end)
    for predicate, args in answer_set:
        yield python_value_to_asp((predicate, args)) + str(atom_end)


def answer_set_from_str(line:str, collapse_atoms:bool=False,
                        collapse_args:bool=True, parse_integer:bool=True) -> iter:
    """Yield atoms found in given string.

    line -- parsable string containing an answer set
    collapse_atoms -- whole atoms are left unparsed
    collapse_args -- atoms args are left unparsed
    parse_integer -- integers are returned as python int objects

    >>> tuple(sorted(answer_set_from_str('a b c d', True)))
    ('a', 'b', 'c', 'd')
    >>> tuple(sorted((answer_set_from_str('a b(a) c("text") d', True))))
    ('a', 'b(a)', 'c("text")', 'd')

    """
    yield from parsing.Parser(
        collapse_atoms=collapse_atoms,
        collapse_args=collapse_args,
        parse_integer=parse_integer
    ).parse_terms(line)


def save_answers_in_file(answers:iter, filename:str or None=None,
                         atom_separator:str=' ',
                         answer_separator:str='\n', end:str='') -> str:
    """Return the name of the file in which the answer sets are written.

    answers -- iterable of answer set to write
    filename -- file to write ; if None, a temporary file will be created
    atom_separator -- string placed between each atom
    answer_separator -- string placed between each answer set
    end -- string written at the end

    """
    if not filename:
        with tempfile.NamedTemporaryFile('w', delete=False) as ofd:
            filename = ofd.name
    with open(filename, 'w') as ofd:
        ofd.write(answer_separator.join(answer_set_to_str(answer, atom_sep=atom_separator)
                                        for answer in answers) + end)
    return filename


def load_answers_from_file(filename:str, answer_set_builder:type=frozenset) -> iter:
    """Yield answer set found in each line of given file"""
    with open(filename) as ifd:
        yield from (
            answer_set_builder(answer_set_from_str(line))
            for line in ifd
        )


def cleaned_path(path:str, error_if_invalid:bool=True) -> str:
    """Return the same path, but cleaned with user expension and absolute"""
    path = os.path.abspath(os.path.expanduser(path))
    if error_if_invalid and not os.path.exists(path):
        open(path)  # will raise FileExistsError
    return path


def with_clingo_bin(clingo_bin:str) -> callable:
    """Generate a wrapper where, during func call, the CLINGO_BIN_PATH clyngor
    variable is set of *clingo_bin*.

    Example:

        @with_clingo_bin('~/bin/clingo-3.2.1')
        def solve_problem(...):
            clyngor.solve(...)

    Not thread safe.

    """
    def wrapper(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            import clyngor
            regular_bin, clyngor.CLINGO_BIN_PATH = clyngor.CLINGO_BIN_PATH, clingo_bin
            ret = func(*args, **kwargs)
            clyngor.CLINGO_BIN_PATH = regular_bin
            return ret
        return wrapped
    return wrapper


def opt_models_from_clyngor_answers(answers:iter):
    """Return generator of optimal models found by clingor.solve from answers.

    This function assumes that:

    - Option '--opt-mode=optN' have been given to clingo.

    - that models are yielded by clingo in increasing optimization value,
    therefore there is no need to verify that a model B succeeding a model A
    is better if they have different optimization value.

    """
    first_seen = False
    optimal_reached = False
    for model, opt, optimum, answer_number in answers.with_answer_number:
        if optimal_reached:
            yield model
        if answer_number == 1 and not first_seen:
            first_seen = True
        elif answer_number == 1:
            optimal_reached = True
            yield model


def try_python_availability_in_clingo(py3=True) -> bool:
    func = try_python_availability_in_clingo_module if clyngor.clingo_module_actived() else try_python_availability_in_clingo_binary
    return func(py3)

def try_python_availability_in_clingo_binary(py3=True) -> bool:
    py_ver = clyngor.clingo_version().get('python')
    if not py_ver:  # NB: python is None, if not available
        return bool(py_ver)
    return py_ver[0] == ('3' if py3 else '2')

def try_python_availability_in_clingo_module(py3=True) -> bool:
    """True if clingo module seems to handle python (3 if py3 is truthy else 2).
    Raise an ModuleNotFoundError if clingo module is not available
    """
    import clingo
    ctl = clingo.Control()
    try:
        ctl.add("base", [], f"#script(python)\n import sys ; assert sys.version.info.major == {'3' if py3 else '2'}\n #end.")
    except RuntimeError as err:  # case where python support is not implemented
        return False
    else:  # python support available
        return True


def try_lua_availability_in_clingo() -> bool:
    return (try_lua_availability_in_clingo_module if clyngor.clingo_module_actived() else try_lua_availability_in_clingo_binary)()

def try_lua_availability_in_clingo_binary() -> bool:
    lua_ver = clyngor.clingo_version().get('lua')
    return bool(lua_ver)  # NB: lua is None, if not available

def try_lua_availability_in_clingo_module() -> bool:
    """True if clingo module seems to handle lua.
    Raise an ModuleNotFoundError if clingo module is not available
    """
    import clingo
    ctl = clingo.Control()
    try:
        ctl.add("base", [], f"#script(lua) #end.")
    except RuntimeError as err:  # case where lua support is not implemented
        return False
    else:  # lua support available
        return True


def null_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(func, *args, **kwargs)
    return wrapper



"""Definition of the Decoder helpers, allowing one to define and use decoders.

A decoder is an object that will be created
when the required atoms are found in the ASP model.

It is heavily inspired from JSON decoders.

"""
from itertools import product
from functools import lru_cache
from inspect import signature
from clyngor import solve


def decode(*args, decoders:iter, **kwargs):
    """Yield objects built from decoders from all models.

    Expects same (kw)args as clyngor.solve, but decoders MUST be provided.

    """
    decoders = tuple(decoders)
    for model in solve(*args, **kwargs).by_predicate:
        for decoder in decoders:
            yield from objects_from_model(decoder, model)

def objects_from_model(decoder:type, model:dict):
    """Yield objects described by given decoder when applied to given model."""
    atoms_group = object_builder_from_decoder(decoder)
    unique_atoms_iterable = tuple(tuple(model.get(atom_name, ())) for atom_name in atoms_group['foreach'])
    all_atoms_iterable = tuple(tuple(model.get(atom_name, ())) for atom_name in atoms_group['all'])
    for unique_atoms in product(*unique_atoms_iterable):
        unique_atoms_args = dict(zip(atoms_group['foreach'], unique_atoms))
        all_atoms_args = dict(zip(atoms_group['all'], all_atoms_iterable))
        yield decoder(**unique_atoms_args, **all_atoms_args)

@lru_cache(maxsize=32)
def object_builder_from_decoder(decoder:type) -> dict:
    if not hasattr(decoder, '__init__'):
        raise ValueError("Decoder {} has no __init__ method".format(decoder))
    constructor = decoder.__init__
    sig = signature(constructor)
    params = iter(sig.parameters.items())
    _ = next(params)  # ignore the self param
    atoms_all, atoms_each = set(), set()
    for name, param in params:
        anno = param.annotation
        if anno is all:  # let's collect all these objects
            atoms_all.add(name)
        elif anno == 1:  # one for each atom
            atoms_each.add(name)
        else:
            raise ValueError("Decoder can't handle the annotation '{}'".format(anno))
    return {
        'all': frozenset(atoms_all),
        'foreach': frozenset(atoms_each),
    }



"""Definition of the easy to integrate ASP code.

"""


from clyngor import solve


def ASP(source_code:str, **kwargs):
    """Return the answer sets obtained by running the solver
    on given raw source code.

    source_code -- ASP source code
    kwargs -- keyword arguments to be given to solve() call

    """
    return solve(inline=source_code, stats=False, **kwargs)


def ASP_last_model(source_code:str, **kwargs):
    """Return the last answer set obtained by running the solver
    on given raw source code.

    source_code -- ASP source code
    kwargs -- keyword arguments to be given to solve() call

    Assuming that given source code is running with an optimization
    and with no --opt-mode option, this function returns an optimal model.

    """
    model = None
    for model in solve(inline=source_code, stats=False, **kwargs):
        pass
    return model


def ASP_one_model(source_code:str, **kwargs):
    """Return the first answer set obtained by running the solver
    on given raw source code.

    source_code -- ASP source code
    kwargs -- keyword arguments to be given to solve() call

    """
    models = solve(inline=source_code, stats=False, nb_model=1, **kwargs)
    return next(models, None)


# shortcuts
ASP.best_model = ASP_last_model
ASP.last_model = ASP_last_model
ASP.one_model = ASP_one_model
ASP.all_models = ASP




from timeit import timeit
from pprint import pprint
from functools import partial

from clyngor import ASP


ASP_CODE = """
rel(a,c). rel(a,d). rel(b,d). rel(b,e).
obj(X):- rel(X,_) ; rel(X,Y): att(Y).
att(Y):- rel(_,Y) ; rel(X,Y): obj(X).
:- not obj(X):obj(X).
:- not att(Y):att(Y).
"""

def time_efficiency(number:int=1000) -> float:
    def run(ASP=ASP, ASP_CODE=ASP_CODE):
        return tuple(ASP(ASP_CODE))
    return 'Perform {} calls in {} seconds.'.format(number, round(timeit(run, number=number), 2))


if __name__ == '__main__':
    answers = ASP(ASP_CODE)
    for answer in answers.by_predicate.first_arg_only:
        print('{' + ','.join(answer['obj']) + '} × {' + ','.join(answer['att']) + '}')
    print()

    print('Benchmark:')
    print(time_efficiency())
    print()



"""Low level interface to clingo.

"""


import re
import os
import json
import shlex
import tempfile
import subprocess
import clyngor
from clyngor.answers import Answers, ClingoAnswers
from clyngor.utils import cleaned_path, ASPSyntaxError, ASPWarning
from clyngor.parsing import parse_clasp_output, validate_clasp_stderr
from clyngor.propagators import Main as _default_running_sequence


def solve(files:iter=(), options:iter=[], inline:str=None,
          decoders:iter=(), subproc_shell:bool=False, print_command:bool=False,
          nb_model:int=0, time_limit:int=0, constants:dict={},
          clean_path:bool=True, stats:bool=True,
          clingo_bin_path:str=None, error_on_warning:bool=False,
          force_tempfile:bool=False, delete_tempfile:bool=True,
          use_clingo_module:bool=True, grounding_observers:iter=(),
          propagators:iter=(), solver_conf:object=None,
          running_sequence:callable=_default_running_sequence,
          programs:iter=(['base', ()],), return_raw_output:bool=False) -> iter:
    """Run the solver on given files, with given options, and return
    an Answers instance yielding answer sets.

    files -- iterable of files feeding the solver
    options -- string or iterable of options for clingo
    inline -- ASP source code to feed the solver with
    decoders -- iterable of decoders to apply on ASP (see clyngor.decoder)
    subproc_shell -- use shell=True in subprocess call (NB: you should not)
    print_command -- print full command to stdout before running it
    clean_path -- clean the path of given files before using them
    stats -- will ask clingo for all stats, instead of just the minimal ones
    clingo_bin_path -- the path to the clingo binary
    error_on_warning -- raise an ASPWarning when encountering a clingo warning
    use_clingo_module -- will use the clingo module (if available)
    force_tempfile -- use tempfile, even if only inline code is given
    delete_tempfile -- delete used tempfiles
    return_raw_output -- don't parse anything, just return iterators over stdout
                         and stderr, without using clingo module

    The following options needs the propagator support and/or python clingo module:
    grounding_observers -- iterable of observers to add to the grounding process
    propagators -- iterable of propagators to add to the solving process
    solver_conf -- clingo.Configuration instance, given to the running_sequence
    running_sequence -- If given, must be a callable taking programs,
                        files and Configuration, returning both clingo.Control
                        and clingo.SolveHandle instances.
    programs -- programs to feed the running sequence with.

    Shortcut to clingo's options:
    nb_model -- number of model to output (0 for all (default), None to disable)
    time_limit -- zero or number of seconds to wait before interrupting solving
    constants -- mapping name -> value of constants for the grounding

    """
    files = [files] if isinstance(files, str) else files
    files = tuple(map(cleaned_path, files) if clean_path else files)
    stdin_feed = None  # data to send to stdin
    use_clingo_module = use_clingo_module and clyngor.have_clingo_module() and not return_raw_output
    if use_clingo_module:
        # the clingo API do not handle stdin feeding
        force_tempfile = True
    if inline and not files and not force_tempfile:  # avoid tempfile if possible
        stdin_feed, inline = inline, None
    elif inline:
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            fd.write(inline)
            tempfile_to_del = fd.name
            files = tuple(files) + (tempfile_to_del,)
            assert files, fd.name
    run_command = command(files, options, inline, nb_model, time_limit,
                          constants, stats, clingo_bin_path=clingo_bin_path)

    if print_command:
        print(run_command)

    if not files and not inline and not stdin_feed:
        # in this case, clingo will wait for stdin input, which will never come
        # so better not call clingo at all, because answer is obvious: one empty model.
        return Answers((), command=' '.join(run_command))


    if use_clingo_module:
        if time_limit != 0 or constants:
            raise ValueError("Options 'time_limit' and 'constants' are not "
                             "handled when used with python clingo module.")
        if solver_conf:
            raise NotImplementedError("Solver configuration handling is currently"
                                      "not implemented")
        options = list(shlex.split(options) if isinstance(options, str) else options)
        ctl = clyngor.clingo_module.Control(options)
        main = running_sequence(programs=programs, files=files, nb_model=nb_model,
                                propagators=propagators, observers=grounding_observers,
                                generator=True)
        return main(ctl)
    else:
        clingo = subprocess.Popen(
            run_command,
            stdin=subprocess.PIPE if stdin_feed else None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=bool(subproc_shell),
        )
        if stdin_feed:
            clingo.stdin.write(stdin_feed.encode())
            clingo.stdin.close()
        stdout = (line.decode() for line in clingo.stdout)
        stderr = (line.decode() for line in clingo.stderr)
        if return_raw_output:  return ''.join(stdout), ''.join(stderr)
        statistics = {}

        # remove the tempfile after the work.
        on_end = lambda: (clingo.stderr.close(), clingo.stdout.close())
        if inline and tempfile_to_del and delete_tempfile:
            on_end = lambda: (os.remove(tempfile_to_del), clingo.stderr.close(), clingo.stdout.close())

        return Answers(_gen_answers(stdout, stderr, statistics, error_on_warning),
                       command=' '.join(run_command), on_end=on_end,
                       decoders=decoders, statistics=statistics,
                       with_optimization=True)


def command(files:iter=(), options:iter=[], inline:str=None,
            nb_model:int=0, time_limit:int=0, constants:dict={},
            stats:bool=True, clingo_bin_path:str=None) -> iter:
    """Return the shell command running the solver on given files,
    with given options.

    files -- iterable of files feeding the solver
    options -- string or iterable of options for clingo
    clingo_bin_path -- the path to the clingo binary

    Shortcut to clingo's options:
    nb_model -- number of model to output (0 for all (default), None to disable)
    time_limit -- zero or number of seconds to wait before interrupting solving
    constants -- mapping name -> value of constants for the grounding
    stats -- True to provides the --stats flag

    """
    files = [files] if isinstance(files, str) else list(files)
    options = list(shlex.split(options) if isinstance(options, str) else options)
    if constants:
        for name, value in constants.items():
            options.append('-c {}={}'.format(name,value))
    if time_limit:
        try:
            time_limit = int(time_limit)
        except ValueError:
            raise ValueError("Time limit should be integer, not " + type(time_limit).__name__)
        options.append('--time-limit=' + str(time_limit))
    if nb_model:
        try:
            nb_model = int(nb_model)
        except ValueError:
            raise ValueError("Number of model must be int, not " + type(nb_model).__name__)
        if nb_model < 0:
            raise ValueError("Number of model must be >= 0.")
        options.append('-n ' + str(nb_model))
    elif nb_model is not None:
        options.append('-n 0')
    if stats:
        options.append('--stats')

    return [clingo_bin_path or clyngor.CLINGO_BIN_PATH, *options, *files]


def clingo_version(clingo_bin_path:str=None) -> dict:
    """Return clingo's version information in a dict"""
    if clyngor.clingo_module_actived():
        return {
            'clingo version': clingo.__version__,
            'python': '3' if clyngor.utils.try_python_availability_in_clingo_module() else None,
            'lua': 'yes' if clyngor.utils.try_lua_availability_in_clingo_module() else None,
        }
    clingo = subprocess.Popen(
        [clingo_bin_path or clyngor.CLINGO_BIN_PATH, '--version', '--outf=2'],
        stderr = subprocess.PIPE,
        stdout = subprocess.PIPE,
    )
    fields = {
        'address model': re.compile(r'Address model: ([3264]{2})-bit'),
        'clingo version': re.compile(r'clingo version ([0-9\.]+)'),
        'libgringo': re.compile(r'libgringo version ([0-9\.]+)'),
        'libclasp': re.compile(r'libclasp version ([0-9\.]+)'),
        'libpotassco': re.compile(r'libpotassco version ([0-9\.]+)'),
        'python': re.compile(r'with[out]{0,3}\sPython\s?([0-9\.]+)?'),  # later loop will yields None if python is available
        'lua': re.compile(r'with[out]{0,3}\sLua\s?([0-9\.]+)?'),  # same for lua
    }
    stdout = clingo.communicate()[0].decode()
    values = {}
    for field, reg in fields.items():
        match = reg.search(stdout)
        values[field] = match.groups()[0] if match else None
    return values



def _gen_answers(stdout:iter, stderr:iter, statistics:dict,
                 error_on_warning:bool) -> (str, int or None, bool, int):
    """Yield 4-uplet (answer set, optimization, optimum found, answer number),
    and update given statistics dict with statistics payloads

    """
    answer = None  # is used to generate a model only when we are sur there is (no) optimization
    answer_number = None
    optimization, optimum_found = None, False
    for ptype, payload in parse_clasp_output(stdout, yield_stats=True):
        if ptype == 'answer_number':
            if answer is not None:
                yield answer, optimization, optimum_found, answer_number
            answer, optimization, optimum_found, answer_number = payload, None, False, None
            answer_number = payload
        elif ptype == 'answer':  # yield previously found answer
            answer = payload
        elif ptype == 'optimum found':
            optimum_found = payload
        elif ptype == 'optimization':
            optimization = payload
        elif ptype == 'statistics':
            statistics.update(payload)
        elif ptype == 'unsat':
            pass  # don't care
        elif ptype == 'info':
            pass  # don't care
        else:
            assert ptype in parse_clasp_output.out_types, 'solving.parse_clasp_output yields an unexpected type ' + repr(ptype)
    if answer is not None:  # if no optimization, probably one miss
        assert answer_number is not None, answer_number
        yield answer, optimization, optimum_found, answer_number

    # handle stderr
    for payload in validate_clasp_stderr(stderr):
        if payload['level'] == 'error' and payload['message'].startswith('syntax error, '):
            raise ASPSyntaxError(
                payload['human message'][len('syntax error, '):],
                (payload['filename'], payload['lineno'],
                 payload['char_beg'], payload['text']),
                payload=payload)
        elif payload['level'] in {'warning', 'info'}:
            if error_on_warning:
                raise ASPWarning(payload['human message'], payload)
            else:
                pass  # do nothing, user said
        else:
            raise SystemError("Clingo yield a non-handled error " + repr(payload))





import pytest
import clyngor
from clyngor.upapi import converted_types
from .definitions import onlyif_python_support


ASP_BASIC = """
#script(python)
from clyngor import converted_types
@converted_types
def h(a:str, b: int):
    return a*b
#end.
r(@h("s", 2)).
"""

ASP_DOUBLE_ARGS = """
#script(python)
from clyngor import converted_types
@converted_types
def g(a:str, b:int) -> tuple:  # the tuple make the return value an iterable of arguments
    yield a * b
    yield len(a) * b
#end.
q(X,Y):- (X,Y)=@g("a",2).
"""

ASP_ALL = """
#script(python)

# Following lines are necessary to use the currently tested clyngor instead
#  of the one installed for the python version used by clingo.
# Note that the 0 index of sys.path is by convention a special one that should
#  not be modified, hence the insert in 1.
import sys
sys.path.insert(1, '.')

from clyngor import converted_types

@converted_types
def f(a:str, b:int, c:int):
    yield a * b
    yield b * c
    yield b + c

@converted_types
def g(a:str, b:int) -> tuple:  # the tuple make the return value an iterable of arguments
    yield a * b
    yield len(a) * b

@converted_types
def h(a:str, b:int) -> 'b'.join:  # any callable will be applied on function output
    for _ in range(b):
        yield a
#end.

p(X):- (X)=@f("hello",2,3).
p(X):- (X)=@f(2,2,3).  % this is a bad type, and therefore ignored

q(X,Y):- (X,Y)=@g("a",2).

r(X):- (X)=@h("a",4).

"""

@onlyif_python_support
def test_all_decorator_usage():
    models = tuple(clyngor.solve(inline=ASP_ALL).by_predicate)
    print(ASP_ALL)
    assert len(models) == 1
    model = models[0]
    assert model == {
        'p': frozenset({('"hellohello"',), (6,), (5,)}),
        'q': frozenset({('"aa"', 2)}),
        'r': frozenset({('"abababa"',)}),
    }

@onlyif_python_support
def test_basic_decorator_usage():
    models = tuple(clyngor.solve(inline=ASP_BASIC).by_predicate)
    print(ASP_BASIC)
    assert len(models) == 1
    model = models[0]
    assert model == {
        'r': frozenset({('"ss"',)}),
    }

@onlyif_python_support
def test_double_args_support():
    models = tuple(clyngor.solve(inline=ASP_DOUBLE_ARGS).by_predicate)
    print(ASP_DOUBLE_ARGS)
    assert len(models) == 1
    model = models[0]
    assert model == {
        'q': frozenset({('"aa"', 2)}),
    }



"""Implementation of decoupled grounding/solving interface"""

from clyngor import solve


def solve_from_grounded(grounded_program:str, **kwargs_to_solver):
    if 'options' not in kwargs_to_solver:
        kwargs_to_solver['options'] = ''
    kwargs_to_solver['options'] += ' --mode=clasp'
    if 'inline' in kwargs_to_solver:
        print('WARNING inline argument was passed to solve_from_grounded. It will be ignored.')
    kwargs_to_solver['inline'] = grounded_program
    return solve(**kwargs_to_solver)


def grounded_program(files:iter=(), inline:str=None, **kwargs_to_solver) -> str:
    "Return full grounded program, ready to be used by solve_from_grounded"
    if 'options' not in kwargs_to_solver:
        kwargs_to_solver['options'] = ''
    kwargs_to_solver['options'] += ' --mode=gringo'
    stdout, stderr = solve(files=files, inline=inline, return_raw_output=True, **kwargs_to_solver)
    return stdout



"""Hooks for pytest.

"""

import pytest

def pytest_configure(config):
    config.addinivalue_line(
        "markers", "slow: this test is slow to execute"
    )

def pytest_addoption(parser):
    parser.addoption('--quick', action='store_true',
                     default=False, help='do not run slow tests')

def pytest_collection_modifyitems(config, items):
    if not config.getoption('--quick'):
        return
    # --quick given in cli: skip slow tests
    skip_slow = pytest.mark.skip(reason="remove --quick option to run")
    for item in items:
        if 'slow' in item.keywords:
            item.add_marker(skip_slow)



CLINGO_BIN_PATH = 'clingo'
__version__ = '0.5.0'

from clyngor.utils import ASPSyntaxError, ASPWarning, parse_clingo_output, clingo_value_to_python, with_clingo_bin, opt_models_from_clyngor_answers, answer_set_to_str, answer_set_from_str
from clyngor.answers import Answers, ClingoAnswers
from clyngor.solving import solve, clingo_version, command
from clyngor.grounding import solve_from_grounded, grounded_program
from clyngor.inline import ASP
from clyngor.decoder import decode
from clyngor.propagators import Propagator, Variable, Main, Constraint


def load_clingo_module() -> bool:
    global clingo_module, clingo_module_available
    try:
        import clingo
        clingo_module = clingo
        clingo_module_available = True
    except ImportError:
        clingo_module = None
        clingo_module_available = False

def have_clingo_module() -> bool:
    return clingo_module is not None

def deactivate_clingo_module():
    global clingo_module
    clingo_module = None

def clingo_module_actived() -> bool:
    return clingo_module is not None

def use_clingo_module():
    if clingo_module is None:
        load_clingo_module()
    if clingo_module is None:  # loading didn't succeed
        raise RuntimeError("Clingo module was asked to be used (call to use_clingo_module), but it is not available.")

def use_clingo_binary(path=None):
    deactivate_clingo_module()
    set_clingo_binary(path or CLINGO_BIN_PATH)

def set_clingo_binary(path):
    globals()['CLINGO_BIN_PATH'] = path

def get_clingo_binary() -> str:
    import shutil
    return shutil.which(globals()['CLINGO_BIN_PATH'])

def have_python_support(py3:bool=True) -> bool or None:
    """True if clingo supports python 3 (or 2 if py3 is falsy).
    None if no python support at all."""
    return utils.try_python_availability_in_clingo(py3)

def have_lua_support() -> bool:
    """True if clingo supports lua"""
    return utils.try_lua_availability_in_clingo(py3)
    return bool(clingo_version()['lua'])


load_clingo_module()  # just initialize clingo module state, whether it is available or not
use_clingo_binary()  # use binary by default
# use_clingo_module()  # use module by default

if CLINGO_BIN_PATH != 'clingo':
    assert get_clingo_binary() == CLINGO_BIN_PATH


# last, clyngor depending modules
from clyngor.upapi import converted_types, converted_types_or_symbols



"""The Answers object"""


import re
from collections import defaultdict

import clyngor
from clyngor import as_pyasp, parsing, utils


def naive_parsing_of_answer_set(answer_set:str, *, discard_quotes:bool=False, parse_int:bool=True, parse_args:bool=True) -> [(str, tuple)]:
    """Yield (pred, args), naively parsed from given answer set encoded as clingo output string.
    Some atoms may be missing. Some others may be poorly parsed."""
    # print('NAIVE_PARSING_OF_ANSWER_SET:', answer_set, f'\t keep_quotes={keep_quotes}, parse_int={parse_int}, parse_args={parse_args}')
    REG_ANSWER_SET = re.compile(r'([a-z_-][a-zA-Z0-9_]*|[0-9]+|"[^"]*")(\([^)]+\))?')

    for match in REG_ANSWER_SET.finditer(answer_set):
        pred, args = match.groups()
        assert args is None or (args.startswith('(') and args.endswith(')'))
        if args and parse_args:
            args = args[1:-1]  # remove surrounding parens
            if discard_quotes:
                args = utils.remove_arguments_quotes(args)
            args = tuple(
                (int(arg) if parse_int and
                 (arg[1:] if arg.startswith('-') else arg).isnumeric() else arg)
                for arg in args.split(',')
            ) if parse_args else args
        elif args:  # args should not be parsed
            args = args
        pred = int(pred) if pred.isnumeric() else pred  # handle
        yield pred, args or ()
        # print('\t>', pred, args)


class Answers:
    """Proxy to the solver, generated by solving methods like solve.solve
    or inline.ASP.

    Iterable on the answer sets generated by the solver.
    Also expose some answer set formatting tunning.

    """

    def __init__(self, answers:iter, command:str='', statistics:dict={},
                 *, decoders:iter=(), with_optimization:bool=False, on_end:callable=None, **kwargs):
        """Answer sets must be iterable of (predicate, args).

        decoders -- iterable of decoders to apply on ASP (see clyngor.decoder).
        with_optimization -- answers are read as ((predicate, args), optimization, optimality)
                             allowing to retrieve optimization data of the answers.
                             See also Answers.with_optimization property.
        on_end -- if callable, called when all answer sets are exhausted.

        """
        if not with_optimization:
            # emulate optimization with None value
            # TODO: verify that in absence of optimization, all answers are ALWAYS enumerated from 1 to n.
            answers = ((answer, None, False, idx) for idx, answer in enumerate(answers, start=1))
        self._answers = iter(answers)
        self._command = str(command or '')
        self._statistics = statistics  # will be updated by reading method
        self._decoders = tuple(decoders)
        self._first_arg_only = False
        self._group_atoms = False
        self._group_atoms_with_arity = False
        self._as_pyasp = False
        self._sorted = False
        self._keep_quotes = True
        # TODO: break api to add keep quotes instead of discard quotes.
        self._careful_parsing = False
        self._collapse_atoms= False
        self._collapse_args = True
        self._parse_int = True
        self._ignore_args = False
        self._with_optimization = False
        self._with_optimality = False
        self._with_answer_number = False
        self.__on_end = on_end or (lambda: None)

        for k, v in kwargs.items():
            if isinstance(getattr(self, '_' + k, None), bool):
                setattr(self, '_' + k, bool(value))
            elif k == 'discard_quotes':
                setattr(self, '_keep_quotes', not bool(value))

    def __del__(self):
        """Call the on_end function, avoiding a too-many-file-open error.

        That error happened in a recent project where clyngor
        was called thousands of times, without accessing all models,
        which prevented __iter__ method to ever call on_end function.
        Hence the use of __del__.

        """
        self.clean_resources()

    @property
    def command(self) -> str:  return self._command

    @property
    def with_optimization(self):
        """Yield (model, optimization) instead of just model"""
        self._with_optimization = True
        return self

    @property
    def with_optimality(self):
        """Yield (model, optimization, optimality) instead of just model"""
        self._with_optimization = True
        self._with_optimality = True
        return self

    @property
    def with_answer_number(self):
        """Yield (model, optimization, optimality, answer_number) instead of just model"""
        self._with_optimization = True
        self._with_optimality = True
        self._with_answer_number = True
        return self

    @property
    def first_arg_only(self):
        """Keep only the first argument, and do not enclose it in a tuple."""
        self._first_arg_only = True
        return self

    @property
    def by_predicate(self):
        """Group atoms by predicate. Answer sets are then dict with predicate
        as keys and collection of args as value."""
        self._group_atoms = True
        return self

    @property
    def by_arity(self):
        """Group atoms by predicate and arity. Answer sets are then dict with
        predicate/arity as keys and collection of args as value."""
        self._group_atoms = True
        self._group_atoms_with_arity = True
        return self

    @property
    def as_pyasp(self):
        """Return Term and TermSet object offering a pyasp-like interface"""
        self._as_pyasp = True
        return self

    @property
    def sorted(self):
        """Sort the atom (or the args when grouped)"""
        self._sorted = True
        return self

    @property
    def keep_quotes(self):
        """Keep the quotes of the string"""
        self._keep_quotes = True
        return self

    @property
    def discard_quotes(self):
        """Discard the quotes of the string"""
        self._keep_quotes = False
        return self

    @property
    def careful_parsing(self):
        """Use robust parser"""
        self._careful_parsing = True
        return self

    @property
    def atoms_as_string(self):
        """All atoms are encoded as ASP strings, left unparsed."""
        self._collapse_atoms = True
        self._collapse_args = True
        return self

    @property
    def int_not_parsed(self):
        """Do not parse the integer arguments, so if an atom have integers
        as arguments, they will be returned as string, not integers.

        """
        self._parse_int = False
        return self

    @property
    def parse_args(self):
        """Parse the arguments as well, so if an atom is argument of another
        one, it will be parsed as any atom instead of being understood
        as a string.

        Will use the robust parser.

        """
        self._careful_parsing = True  # needed to implement the collapse
        self._collapse_atoms = False
        self._collapse_args = False
        return self

    @property
    def no_arg(self):
        """Do not parse arguments, and discard/ignore them.

        """
        self._ignore_args = True
        return self


    def __next__(self):
        return next(iter(self))

    def __iter__(self):
        """Yield answer sets"""
        for answer_set, optimization, optimality, answer_number in self._answers:
            answer_set = tuple(self._parse_answer(answer_set))
            parsed = self._format(answer_set)
            if self._with_answer_number:
                yield parsed, optimization, optimality, answer_number
            elif self._with_optimality:
                yield parsed, optimization, optimality
            elif self._with_optimization:
                yield parsed, optimization
            else:
                yield parsed
        self.clean_resources()

    def clean_resources(self):
        self.__on_end()
        self.__on_end = lambda: None  # don't call it again


    def _parse_answer(self, answer_set:str) -> iter:
        """Yield atoms as (pred, args) from given answer set"""
        careful_parsing = self._careful_parsing or parsing.careful_parsing_required(answer_set)
        # keep_quotes = self._keep_quotes or not self._collapse_atoms or not self._as_pyasp
        discard_quotes = not (self._keep_quotes or self._collapse_atoms or self._as_pyasp)
        if isinstance(answer_set, str) and careful_parsing:
            # print('CAREFUL PARSING:', answer_set)
            # _keep_quotes is incompatible with atoms_as_string and as_pyasp.
            # atom_as_string: remove the quotes delimiting arguments.
            # as_pyasp: remove the quotes for the arguments.
            yield from parsing.Parser(
                self._collapse_atoms, self._collapse_args,
                discard_quotes,
                self._first_arg_only,
                parse_integer=self._parse_int
            ).parse_terms(answer_set)
        elif isinstance(answer_set, str):  # the good ol' split
            # print('THE GOOD OLD SPLIT:', f"keep_quotes={self._keep_quotes}  collapse_atoms={self._collapse_atoms}")
            yield from self.__finish_parsing(naive_parsing_of_answer_set(answer_set, discard_quotes=discard_quotes, parse_int=self._parse_int, parse_args=self._collapse_args or self._first_arg_only))
        elif isinstance(answer_set, (set, tuple)) and all(isinstance(atom, (str, int, tuple)) for atom in answer_set):  # already parsed
            # print('FROM SET OR TUPLE')
            if not self._parse_int:
                answer_set = utils.integers_to_string_atoms(answer_set)
            yield from self.__finish_parsing(answer_set)
        else:  # unknown format
            raise ValueError("unknow answer set format: {}, {}".format(type(answer_set), answer_set))

    def __finish_parsing(self, answer_set:[(str, tuple)]) -> [(str, tuple) or str]:
        """Modify (pred, args) atoms according to parsing options"""
        for pred, args in answer_set:
            if self._first_arg_only:
                args = [args[0]] if args else args
            if self._atoms_as_string:
                args = ('(' + ','.join(map(str, args)) + ')') if args else ''
                yield pred + args
            else:
                yield pred, args


    def _format(self, answer_set) -> dict or frozenset:
        """Perform the formatting of the answer set according to
        formatting options.

        answer_set -- iterable of (pred, args)

        """
        sorted_tuple = lambda it: tuple(sorted(it))
        builder = sorted_tuple if self._sorted else frozenset
        if self._atoms_as_string:  # special case
            return builder(answer_set)
        elif self._ignore_args:
            answer_set = (pred for pred, _ in answer_set)
            if self._group_atoms:
                return {pred: frozenset() for pred in answer_set}
            if self._as_pyasp:
                return as_pyasp.TermSet(as_pyasp.Atom.from_tuple_repr((pred, ())) for pred in answer_set)
            return builder(answer_set)
        elif self._first_arg_only:
            answer_set = builder((pred, args[0] if args else ())
                                   for pred, args in answer_set)
        else:
            answer_set = builder((pred, tuple(args))
                                   for pred, args in answer_set)
        # NB: as_pyasp flag behave differently if group_atoms is activated
        if self._group_atoms:
            mapping = defaultdict(set)
            for pred, args in answer_set:
                if self._as_pyasp:
                    args = as_pyasp.Atom.from_tuple_repr((pred, args))
                mapping[pred].add(args)
                if self._group_atoms_with_arity:
                    mapping[pred, len(args)].add(args)
                    mapping['{}/{}'.format(pred, str(len(args)))].add(args)
            if self._as_pyasp:
                return {pred: as_pyasp.TermSet(args) for pred, args in mapping.items()}
            else:
                return {pred: builder(args) for pred, args in mapping.items()}
        elif self._as_pyasp:
            return as_pyasp.TermSet(as_pyasp.Atom.from_tuple_repr(atom) for atom in answer_set)
        return answer_set


    @property
    def _atoms_as_string(self) -> bool:
        """Shortcut"""
        return self._collapse_atoms and self._collapse_args

    @property
    def statistics(self) -> dict:
        return dict(self._statistics)


class ClingoAnswers(Answers):
    """Proxy to the solver as called through the python clingo module.

    """
    def __init__(self, solver, statistics:callable=(lambda: {})):
        assert clyngor.have_clingo_module()
        super().__init__(self.__compute_answers(), with_optimization=True, command='[clingo module call]')
        self._solver = solver
        self._statistics = lambda s=solver: s.statistics
        assert callable(self._statistics)


    def __compute_answers(self):
        kwargs = {'yield_': True, 'async_': True}  # compat with 3.7
        with self._solver.solve(**kwargs) as models:
            for model in models:
                answer_set = set(utils.clingo_symbol_as_python_value(a)
                                 for a in model.symbols(shown=True))
                yield answer_set, model.cost, model.optimality_proven, model.number

    @property
    def statistics(self) -> dict:
        return self._statistics()



"""Stress tests for clyngor.
"""

import pytest
from clyngor import solve
from .definitions import run_with_clingo_binary_only



@pytest.mark.slow
@run_with_clingo_binary_only
def test_many_call_to_test_file_closing():
    """at some point, if files are not closed, this should fail"""
    for idx in range(100):  # arbitrary number of call leading to an error
        models = solve(inline=f'a({idx}).', delete_tempfile=False, force_tempfile=True)  # force the use of tempfile
        assert next(models) == {('a', (idx,))}
        with pytest.raises(StopIteration):
            next(models)



"""Routines for answer set parsing.

"""
import re


import arpeggio as ap


class CollapsableAtomVisitor(ap.PTNodeVisitor):
    """Implement both the grammar and the way to handle it, dedicated to the
    parsing of ASP like string to produce frozenset instances.

    collapse_args: function terms in predicate arguments are collapsed into strings
    collapse_atoms: atoms (predicate plus terms) are collapsed into strings
                   requires that collapse_args is True

    See Parser class for usage examples.

    """
    def __init__(self, collapse_args:bool=True, collapse_atoms:bool=False,
                 discard_quotes:bool=False, first_arg_only:bool=False, parse_integer:bool=True):
        super().__init__()
        self.collapse_args = bool(collapse_args)
        self.collapse_atoms = bool(collapse_atoms)
        self._int_builder = int if parse_integer else str
        self.discard_quotes = bool(discard_quotes)
        self.first_arg_only = bool(first_arg_only)

    def visit_number(self, node, children):
        return self._int_builder(node.value)

    def visit_args(self, node, children):
        return children

    def visit_single_value(self, node, children):
        return self.visit_subterm(node, children)
    def visit_one_uplet(self, node, children):
        return self.visit_atom(node, ('', children))
    def visit_n_uplet(self, node, children):
        return self.visit_atom(node, ('', children))

    def visit_text(self, node, children):
        text = tuple(children)[0] if children else ''
        if self.discard_quotes:
            return text
        else:
            return '"' + text + '"'

    def visit_subterm(self, node, children):
        predicate, *args = children
        if self.collapse_args:
            return (predicate + '(' + ','.join(map(str, *args)) + ')') if args else predicate
        else:
            return (predicate, tuple(args[0])) if args else predicate

    def visit_atom(self, node, children):
        predicate, *args = children

        if self.first_arg_only and args:
            args = [[arg[0] for arg in args]]

        if self.collapse_atoms:
            return (predicate + '(' + ','.join(map(str, *args)) + ')') if args else predicate
        else:
            return (predicate, (tuple(args[0]) if args else ()))

    def visit_terms(self, node, children):
        return frozenset(children)

    @staticmethod
    def grammar():
        def ident():      return ap.RegExMatch(r'[a-z_-][a-zA-Z0-9_]*')
        def number():     return ap.RegExMatch(r'-?[0-9]+')
        def text():       return '"', ap.RegExMatch(r'((\\")|([^"]))*'), '"'
        def subterm():    return [number, text, (ident, ap.Optional('(', args, ')')), aloneargs]
        def args():       return subterm, ap.ZeroOrMore(',', subterm), ap.Optional(',')
        def aloneargs():  return [single_value, one_uplet, n_uplet]
        def single_value():  return '(', subterm, ')'
        def one_uplet():  return '(', subterm, ',', ')'
        def n_uplet():    return '(', subterm, ap.OneOrMore(',', subterm), ')'
        # NB: litteral outputed by #show are not handled.
        def atom():       return ident, ap.Optional("(", args, ")")
        def term():       return [number, text, atom]
        def terms():      return ap.ZeroOrMore(term)
        return terms


class Parser:
    def __init__(self, collapse_atoms=False, collapse_args=True, discard_quotes:bool=False,
                 first_arg_only:bool=False, callback=None, parse_integer:bool=True):
        """
        collapse_args -- function terms in predicate arguments are collapsed into strings
        collapse_atoms -- atoms (predicate plus terms) are collapsed into strings
                          requires that collapse_args is True
        parse_integer -- return integers as int instead of string

        examples:

        >>> Parser(False, True).parse_terms('a(b,c(d))')
        frozenset({('a', ('b', 'c(d)'))})

        >>> Parser(True, True).parse_terms('a(b,c(d))')
        frozenset({'a(b,c(d))'})

        >>> Parser(False, False).parse_terms('a(b,c(d))')
        frozenset({('a', ('b', ('c', ('d',))))})

        >>> Parser(False, False).parse_terms('a')
        frozenset({('a', ())})

        >>> Parser(True, False).parse_terms('a(b,c(d))')  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        ValueError

        >>> Parser(parse_integer=False).parse_terms('a(-3)')
        frozenset({('a', ('-3',))})

        >>> Parser(parse_integer=True).parse_terms('a(-2)')
        frozenset({('a', (-2,))})

        >>> Parser(discard_quotes=True).parse_terms('a("b")')
        frozenset({('a', ('b',))})
        >>> Parser().parse_terms('a("b")')
        frozenset({('a', ('"b"',))})

        """
        self.collapse_args = bool(collapse_args)
        self.collapse_atoms = bool(collapse_atoms)
        self.discard_quotes = bool(discard_quotes)
        self.first_arg_only = bool(first_arg_only)
        self.atom_visitor = CollapsableAtomVisitor(
            self.collapse_args,
            self.collapse_atoms,
            self.discard_quotes,
            self.first_arg_only,
            parse_integer
        )
        self.grammar = self.atom_visitor.grammar()
        self.callback = callback
        if self.collapse_atoms and not self.collapse_args:
            raise ValueError("if atoms are collapsed, terms must"
                             " also be collapsed!")


    def parse_terms(self, string:str) -> frozenset:
        """Return the frozenset computed from given valid ASP-compliant string"""
        parse_tree = ap.ParserPython(self.grammar).parse(string)
        if parse_tree:
            return ap.visit_parse_tree(parse_tree, self.atom_visitor)
        else:
            return frozenset()

    # alias
    parse = parse_terms


    def parse_clasp_output(self, output:iter or str, *, yield_stats:bool=False,
                           yield_info:bool=False, yield_prgs:bool=False):
        """Decorator over the parse_clasp_output module function,
        where the answers are parsed using self.parse_terms method.

        """
        parsed = parse_clasp_output(output, yield_stats=yield_stats,
                                    yield_info=yield_info,
                                    yield_prgs=yield_prgs)
        for type, payload in parsed:
            if type == 'answer':
                yield type, self.parse_terms(payload)
            else:
                yield type, payload


def parse_clasp_output(output:iter or str, *, yield_stats:bool=False,
                       yield_opti:bool=True, yield_info:bool=False,
                       yield_prgs:bool=False):
    """Yield pairs (payload type, payload) where type is 'info', 'statistics',
    'optimization', 'progression', or 'answer' and payload the raw information.

    output -- iterable of lines or full clasp output to parse
    yield_stats -- yields final statistics as a mapping {field: value}
                   under type 'statistics'
    yield_opti  -- yields line sometimes following an answer set,
                   beginning with 'Optimization: ' or 'OPTIMUM FOUND'.
    yield_info  -- yields all lines not included in other types, including the
                   first lines not related to first answer
                   under type 'info' as a tuple of lines
    yield_prgs -- yields lines of Progressions that are sometimes
                  following an answer set in multithreading contexts,
                  under type 'progression' as a string.

    In any case, tuple ('answer', termset) will be returned
    with termset a string containing the raw data.

    """
    ASW_FLAG, OPT_FLAG, OPT_FOUND, PROGRESS = 'Answer: ', 'Optimization: ', 'OPTIMUM FOUND', 'Progression :'
    UNSAT = 'UNSATISFIABLE'
    output = iter(output.splitlines() if isinstance(output, str) else output)

    # get the first lines
    line = next(output)
    infos = []
    while not line.startswith(ASW_FLAG) and not line.startswith(UNSAT):
        infos.append(line)
        try:
            line = next(output)
        except StopIteration:
            return

    # first answer begins
    while True:
        if line.startswith(ASW_FLAG):
            answer_number = line[len(ASW_FLAG):].split(" ")[0]  
            yield 'answer_number', int(answer_number)
            yield 'answer', next(output)
        elif line.startswith(OPT_FLAG) and yield_opti:
            yield 'optimization', tuple(map(int, line[len(OPT_FLAG):].strip().split()))
        elif line.startswith(OPT_FOUND) and yield_opti:
            yield 'optimum found', True
        elif line.startswith(PROGRESS) and yield_prgs:
            yield 'progression', line[len(PROGRESS):].strip()
        elif line.startswith(UNSAT):
            yield 'unsat', True
        elif not line.strip():  # empty line: statistics are beginning
            if not yield_stats: break  # stats are the last part of the output
            stats = {}
            for line in output:
                sep = line.find(':')
                key, value = line[:sep], line[sep+1:]
                stats[key.strip()] = value.strip()
            yield 'statistics', stats
            break
        else:  # should not happen
            infos.append(line)
        try:
            line = next(output)
        except StopIteration:
            break

    if yield_info:
        yield 'info', tuple(infos)

parse_clasp_output.out_types = ('info', 'answers', 'optimization, ''statistics')  # the order is the one in clingo input


def validate_clasp_stderr(stderr:iter or str) -> iter:
    """Parse stderr of clingo, detect and yield defects lines in form of dict"""
    reg_err = re.compile(r'(.+):([0-9]+):([0-9]+)-([0-9]+): (\w+): (.+)')
    while True:
        try:
            line = next(stderr).strip()
        except StopIteration:
            return
        err_match = reg_err.fullmatch(line)
        if err_match:
            data = dict(zip(('filename', 'lineno', 'char_beg', 'char_end', 'level', 'message'), err_match.groups()))
            for int_field in ('lineno', 'char_beg', 'char_end'):
                data[int_field] = int(data[int_field])
            data['text'] = line
            data['human message'] = '{} in file {} at line {} and column {}-{}'.format(
                data['message'].strip(':,'), data['filename'], data['lineno'], data['char_beg'], data['char_end']
            )

            # special cases
            if 'atom does not occur in any rule head' in data['text']:
                data['atom'] = next(stderr).strip()
                data['human message'] = "atom '{}' does not occur in any rule head in file {} at line {} and column {}-{}".format(
                    data['atom'], data['filename'], data['lineno'], data['char_beg'], data['char_end']
                )
            yield data


def careful_parsing_required(answer_set:str) -> bool:
    """True if given answer set as string requires to be carefully parsed.
    Note that this function has no false positive, but numerous false negative.
    """
    in_paren = False
    in_quotes = False
    escape = False
    for char in answer_set:
        if char == '(':
            if in_paren or in_quotes:
                return True  # paren in paren (or quotes) requires careful parsing
            in_paren = True
        elif char == ')':
            in_paren = False
        elif char == '"':
            if escape:
                escape = False
                continue
            in_quotes = not in_quotes
        elif char == ',':
            if in_quotes:
                return True  # comma in quotes requires careful parsing
        elif char == '\\':
            escape = True
    return False




import tempfile
import pytest
import clyngor
from clyngor import ASP, solve, command
from clyngor import utils, CLINGO_BIN_PATH
from .definitions import run_with_clingo_binary_only


@pytest.fixture
def asp_code():
    return """
    rel(a,(c;d)). rel(b,(d;e)).
    obj(X):- rel(X,_) ; rel(X,Y): att(Y).
    att(Y):- rel(_,Y) ; rel(X,Y): obj(X).
    :- not obj(X):obj(X).
    :- not att(Y):att(Y).
    """


def test_api_solve_with_weird_flags():
    files = []
    with tempfile.NamedTemporaryFile('wt', delete=False) as fd:
        fd.write('1 {a ; b ; c} 1 :- d.')
        files.append(fd.name)
    with tempfile.NamedTemporaryFile('wt', delete=False) as fd:
        fd.write('0 {d ; e} 1.')
        files.append(fd.name)

    answers = solve(files, options='--parallel-mode=4').no_arg
    print(answers.command)
    set = frozenset
    assert set(answers) == { set('da'), set('db'), set('dc'), set('e'), set() }

    answers = solve(files, options='--opt-mode=optN').no_arg
    print(answers.command)
    set = frozenset
    assert set(answers) == { set('da'), set('db'), set('dc'), set('e'), set() }

    answers = solve(files, options='--parallel-mode=4 --opt-mode=optN').no_arg
    print(answers.command)
    set = frozenset
    assert set(answers) == { set('da'), set('db'), set('dc'), set('e'), set() }


def test_api_solve():
    files = []
    with tempfile.NamedTemporaryFile('wt', delete=False) as fd:
        fd.write('1 {a ; b ; c} 1 :- d.')
        files.append(fd.name)
    with tempfile.NamedTemporaryFile('wt', delete=False) as fd:
        fd.write('0 {d ; e} 1.')
        files.append(fd.name)
    answers = solve(files).no_arg
    print(answers.command)
    set = frozenset
    assert set(answers) == { set('da'), set('db'), set('dc'), set('e'), set() }


def test_api_command():
    files = ('a.lp', 'b.lp')
    cmd = command(files, nb_model=3, stats=False)
    assert cmd == [CLINGO_BIN_PATH, '-n 3', *files]

    cmd = command(files, nb_model=None)
    assert cmd == [CLINGO_BIN_PATH, '--stats', *files]

    files = ('a.lp', 'b.lp', 'c')
    clyngor.CLINGO_BIN_PATH = '/usr/bin/clingo'  # NB: this have serious side effects. If any fail happen before the restauration, all other tests may fail.
    cmd = command(files, nb_model=0)
    clyngor.CLINGO_BIN_PATH = 'clingo'
    assert cmd == ['/usr/bin/clingo', '-n 0', '--stats', *files]


def test_api_asp(asp_code):
    answers = ASP(asp_code, use_clingo_module=False)  # clingo module offers a *really* different statistics set
    found = set()
    for answer in answers.by_predicate.sorted.first_arg_only:
        found.add(''.join(answer['obj']) + '×' + ''.join(answer['att']))
    assert found == {'a×cd', 'b×de', 'ab×d'}
    assert len(answers.statistics) == 4
    assert answers.statistics['Calls'] == '1'
    assert answers.statistics['Models'] == '3'
    assert set(answers.statistics.keys()) == {'CPU Time', 'Calls', 'Models', 'Time'}


def test_api_inline_by_solve(asp_code):
    answers = solve((), inline=asp_code)
    found = set()
    for answer in answers.by_predicate.sorted.first_arg_only:
        found.add(''.join(answer['obj']) + '×' + ''.join(answer['att']))
    assert found == {'a×cd', 'b×de', 'ab×d'}


def test_string_with_lot_of_crap():
    dangerous_string = r'"\"1,3-dimethyl-2-[2-oxopropyl thio]imidazolium chloride\""'
    for sol in clyngor.ASP('atom({}).'.format(dangerous_string)).careful_parsing:
        assert len(sol) == 1
        pred, args = next(iter(sol))
        assert pred == 'atom'
        assert len(args) == 1
        assert args[0] == dangerous_string


def test_string_without_escaped_quotes():
    dangerous_string = r'1,3-dimethyl-2-[2-oxopropyl thio]imidazolium chloride'
    quoted_dangerous_string = f'"{dangerous_string}"'
    for sol in clyngor.ASP('atom({}).'.format(quoted_dangerous_string)).careful_parsing:
        assert len(sol) == 1
        pred, args = next(iter(sol))
        assert pred == 'atom'
        assert len(args) == 1
        assert args[0] == quoted_dangerous_string


def test_lispy_construct():
    dangerous_string = r'1,3-dimethyl-2-[2-oxopropyl thio]imidazolium chloride'
    quoted_dangerous_string = f'"{dangerous_string}"'
    source = r'atom((((a,b),c),d),{},(1,(2,3))). yield(X,Y,Z):- atom(X,Y,Z).'.format(quoted_dangerous_string)
    for sol in clyngor.ASP(source).careful_parsing.by_predicate.parse_args:
        assert len(sol) == 2
        assert sol['atom'] == sol['yield']
        args = next(iter(sol['yield']))
        assert len(args) == 3
        assert args[0] == ('', (('', (('', ('a', 'b')), 'c')), 'd'))
        assert args[1] == quoted_dangerous_string
        assert args[2] == ('', (1, ('', (2, 3))))


def test_no_input(capsys):
    """This test ensure that empty input do not lead the code to wait forever
    clingo to finish, since it will just wait for stdin input.

    So if this test takes time to run, you know why.

    Note the use of capsys fixture, allowing to disable the input patching
    that make it fail automatically, and therefore disturbing the functionality.

    """
    with capsys.disabled():
        assert tuple(clyngor.solve((), inline='')) == ()


@run_with_clingo_binary_only
def test_syntax_error():
    assert not clyngor.have_clingo_module()
    with pytest.raises(clyngor.ASPSyntaxError) as excinfo:
        tuple(clyngor.solve((), inline='invalid', force_tempfile=True))
    assert excinfo.value.filename.startswith(tempfile.gettempdir())
    assert excinfo.value.lineno == 2
    assert excinfo.value.offset == 1
    assert excinfo.value.payload['char_end'] == 2
    assert excinfo.value.msg.startswith(f'unexpected EOF in file {tempfile.gettempdir()}')
    assert excinfo.value.msg.endswith(' at line 2 and column 1-2')


@run_with_clingo_binary_only
def test_syntax_error_semicolon():
    with pytest.raises(clyngor.ASPSyntaxError) as excinfo:
        tuple(clyngor.solve((), inline='color(X,red):- ;int(X,"adult").', force_tempfile=True))
    assert excinfo.value.filename.startswith(tempfile.gettempdir())
    assert excinfo.value.lineno == 1
    assert excinfo.value.offset == 16
    assert excinfo.value.msg.startswith(f'unexpected ; in file {tempfile.gettempdir()}')
    assert excinfo.value.msg.endswith(' at line 1 and column 16-17')


@run_with_clingo_binary_only
def test_syntax_error_brace():
    with pytest.raises(clyngor.ASPSyntaxError) as excinfo:
        tuple(clyngor.solve((), inline='color(X,red):- {{}}.', force_tempfile=True))
    assert excinfo.value.filename.startswith(tempfile.gettempdir())
    assert excinfo.value.lineno == 1
    assert excinfo.value.offset == 17
    assert excinfo.value.msg.startswith(f'unexpected {{ in file {tempfile.gettempdir()}')
    assert excinfo.value.msg.endswith(' at line 1 and column 17-18')


@run_with_clingo_binary_only
def test_syntax_error_brace_with_stdin():
    with pytest.raises(clyngor.ASPSyntaxError) as excinfo:
        tuple(clyngor.solve((), inline='color(X,red):- {{}}.'))
    assert excinfo.value.filename == '-'
    assert excinfo.value.lineno == 1
    assert excinfo.value.offset == 17
    assert excinfo.value.msg.startswith('unexpected { in file -')
    assert excinfo.value.msg.endswith(' at line 1 and column 17-18')


@run_with_clingo_binary_only
def test_undefined_warning():
    assert not clyngor.have_clingo_module()
    with pytest.raises(clyngor.ASPWarning) as excinfo:
        tuple(clyngor.solve((), inline='b:- c.', error_on_warning=True, force_tempfile=True))
    assert excinfo.value.atom == 'c'
    assert len(excinfo.value.args) == 1
    start = f"atom 'c' does not occur in any rule head in file {tempfile.gettempdir()}"
    assert excinfo.value.args[0].startswith(start)
    assert excinfo.value.args[0].endswith(" at line 1 and column 5-6")

    # NB: the following should NOT raise any error (default value)
    tuple(clyngor.solve((), inline='b:- c.', error_on_warning=False))
    tuple(clyngor.solve((), inline='b:- c.'))


def test_default_negation():
    "Will fail if official clingo module is used and default negation not handled"
    CODE = """
    p(1..2).
    q(2..4).

    -r(X):- not r(X), p(X).

    #show r/1.
    #show -r/1.
    """
    models = clyngor.solve(inline=CODE)
    model = next(models, None)
    assert next(models, None) is None
    print(model)
    assert model == {('-r', (1,)), ('-r', (2,))}


def test_unsatisfiable():
    "Should return an empty answers set"
    CODE = """
    p(1).
    q(1).

    :- p(X), q(X).
    """
    models = clyngor.solve(inline=CODE, stats=False)
    model = next(models, None)
    assert model is None
    assert len(models.statistics) == 4


def test_unsatisfiable_statistics():
    "Should return an empty answers set but provide the statistics"
    CODE = """
    p(1).
    q(1).

    :- p(X), q(X).
    """
    models = clyngor.solve(inline=CODE, stats=True)
    model = next(models, None)
    assert model is None
    assert len(models.statistics) > 4



"""Definitions for propagators API of clyngor.

Exposes Main, Constraint and Variable.

"""

import os
import math
import tempfile
from collections import defaultdict
from . import utils
from .answers import ClingoAnswers


def Main(files:iter=(), inline:str='', nb_model:int=0,
         propagators:iter or object=(), observers:iter or object=(),
         programs:iter or dict={'base': ()}, generator:bool=False):
    """Main function builder for clingo.

    Allow user to use:

        from clyngor import Main
        main = Main()

    instead of rewriting the standard main function:

        def main(prg):
            prg.ground([('base', [])])
            return prg.solve()

    Also, Main() will add to the function additional treatments based on:

    propagators -- propagators instances to register (default none)
    observers -- observers instances to register (default none)
    programs -- programs with their args to ground (default base without param)
    files -- list of files to ground
    inline -- ASP code to add to base program
    generator -- the main function will return a ClingoAnswers instance instead
                 of returning the solve call result
    nb_model -- number of model to search for. 0 stands for all.

    """
    if not isinstance(propagators, (tuple, list, set, frozenset)):
        propagators = (propagators,)
    if not isinstance(observers, (tuple, list, set, frozenset)):
        observers = (observers,)

    programs = list(
        (prg, list(args)) for prg, args
        in (programs.items() if isinstance(programs, dict) else programs)
    )
    if inline:
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            fd.write(inline)
            tempfile_to_del = fd.name
            files = tuple(files) + (tempfile_to_del,)
            assert files, fd.name

    def main(prg):
        for file in files:
            prg.load(file)
        if inline and tempfile_to_del:  # remove the tempfile after the work.
            os.remove(tempfile_to_del)
        for observer in observers:
            prg.register_observer(observer)
        for propagator in propagators:
            prg.register_propagator(propagator)
        prg.ground(programs)
        prg.configuration.solve.models = nb_model
        if generator:
            return ClingoAnswers(prg)  # will take care of solving by itself
        else:
            prg.solve()
    return main


# Placeholder value for 'any value' in input atoms of Constraint's formulas.
Variable = Ellipsis


class Propagator:
    """Base class for high level propagator."""

    def __init__(self, follow:iter or str=()):
        """
        follow -- iterable or string describing the atoms to follow during solving
        """
        self._followeds = frozenset(utils.make_hashable(follow))
        self._str_followeds = frozenset(fol for fol in self._followeds
                                        if isinstance(fol, str))
        self._raw_followeds = frozenset(fol for fol in self._followeds
                                        if not isinstance(fol, str))
        self.__discarding_model = False


    def init(self, init):
        self.symbols = defaultdict(set)
        if not self._followeds: return
        for atom in init.symbolic_atoms:
            lit = init.solver_literal(atom.literal)
            repr_str = str(atom.symbol)
            if self._match_str(repr_str):
                init.add_watch(lit)
                self.symbols[lit].add(repr_str)
            else:  # not in string inputs, maybe in raws ?
                repr = atom.symbol.name, tuple(map(utils.clingo_value_to_python,
                                                   atom.symbol.arguments))
                if self._match_raw(*repr):
                    init.add_watch(lit)
                    self.symbols[lit].add(repr)


    def propagate(self, ctl, changes):
        if self.__discarding_model:
            # propagation during model discarding
            return
        value_of = ctl.assignment.value
        values = {
            repr: value_of(lit)
            for lit, reprs in self.symbols.items()
            for repr in reprs
        }
        complete = all(value is not None for value in values.values())
        partial = any(value is not None for value in values.values())
        if complete and hasattr(self, 'on_all_input'):
            discard = self.on_all_input(values)
        elif partial and hasattr(self, 'on_any_input'):
            discard = self.on_any_input(values)
        else:
            return

        if discard:
            self.__discard_model(ctl, use_lit=next(iter(changes)))


    def __discard_model(self, ctl, use_lit):
        """Make the current model false by adding a clause
        """
        # make it true, then false.
        self.__discarding_model = True
        add_clause_failure = not ctl.add_nogood([use_lit], tag=True, lock=True)
        if add_clause_failure:
            # print('STATE:', add_clause_failure)
            self.__discarding_model = False
            return
        propagation_failure = not ctl.propagate()
        self.__discarding_model = False
        # print('STATE:', add_clause_failure, propagation_failure)
        if add_clause_failure or propagation_failure:
            return


    def _match_str(self, atom:str) -> bool:
        """True if given string atom is in inputs"""
        if atom in self._str_followeds:
            return True

    def _match_raw(self, name:str, args:tuple) -> bool:
        """True if given atom is in inputs"""
        if (name, args) in self._raw_followeds or (name,) in self._raw_followeds:
            return True
        # more complex verification: it may be because of variables
        for atom in self._raw_followeds:
            if len(atom) != 2: continue
            pred, params = atom
            if pred == name and len(args) == len(params):
                for arg, param in zip(args, params):
                    if param is Variable or arg == param:
                        return True


    def run_with(self, filenames:[str]=(), inline:str='',
                 programs:iter=(['base', ()],), options:list=[]):
        import clingo
        ctl = clingo.Control(options)
        main = Main(propagators=self, programs=programs, inline=inline, generator=True)
        return main(ctl)


class Constraint(Propagator):
    """Base class for a particular class of user defined propagators.

    Instances of PyConstraint are valid propagators.
    The basic use is to provides the constructor with a callable
    and the set of atoms to watch and pass to the said callable.

    >>> Constraint(lambda ins: ins['q'], inputs={'q'})  #doctest: +ELLIPSIS
    <clyngor.propagators.Constraint object at ...>

    """

    def __init__(self, formula:callable, inputs:[tuple]):
        super().__init__(follow=inputs)
        self.__formula = formula

    def on_all_input(self, values:dict):
        return self.__formula(values)

    # def on_all_inputs(self):
        # pass





"""The following definitions are here to provides to end-user
a pyasp-like interface, facilitating migrations towards it.

Note that all the interface is not available. Just the very basics.

"""

import tempfile
import itertools

class Atom:
    """Keep an interface close to pyasp"""

    def __init__(self, predicate:str, args:iter=()):
        self.predicate = predicate
        self.arguments = tuple(args)

    def args(self): return self.arguments

    def nb_args(self):
        return len(self.arguments)
    def arg(self, n):
        return self.arguments[n]

    def __eq__(self, other_atom)-> bool:
        return self.predicate == other_atom.predicate and self.arguments == other_atom.arguments

    def __hash__(self):
        return hash((self.predicate, self.arguments))

    def __str__(self):
        if self.nb_args() > 0:
            return '{}({})'.format(self.predicate, ','.join(map(str, self.arguments)))
        else:
            return self.predicate

    def __repr__(self): return str(self)

    def __iter__(self):
        """Conserve the same API as regular representation of atoms"""
        return iter((self.predicate, self.arguments))

    def from_tuple_repr(atom:tuple) -> object:
        "Expects something like ('a', (('b', ('c', 'd')))), encoding a(b(c,d))"
        def from_args(args:tuple) -> object:
            for arg in args:
                if isinstance(arg, (tuple, list)):
                    yield Atom.from_tuple_repr(arg)
                else:
                    yield arg
        pred, args = atom
        return Atom(pred, from_args(args))



# aliases used by pyasp
Term = Atom

class TermSet:
    """Collection of Term/Atom.

    Behave (about) like a set.

    """

    def __init__(self, terms:iter=()):
        self._terms = frozenset(terms)

    def __iter__(self):
        return iter(self._terms)

    def __len__(self):
        return sum(1 for term in self._terms)

    def __eq__(self, other):
        return set(self._terms) == other

    def add(self, atom:Atom):
        "Add given Atom to self"
        self._terms |= frozenset({atom})

    def union(self, *others:object) -> object:
        "Return a new TermSet containing all Atoms of self and each of others TermSet"
        return TermSet(itertools.chain(self._terms, *(other._terms for other in others)))

    def to_file(self, fname:str=None):
        """Return the filename in which term set have been written"""
        if not fname:
            with tempfile.NamedTemporaryFile('w', delete=False) as fd:
                fname = fd.name
        with open(fname, 'w') as fd:
            for term in self:
                fd.write(str(term) + '.\n')
        return fname

    def __str__(self):
        return '.'.join(map(str, self)) + ('.' if self._terms else '')

    def __repr__(self): return str(self)



import pytest
from clyngor import solve
from .definitions import run_with_clingo_binary_only


@pytest.mark.slow
@run_with_clingo_binary_only
def test_time_limit_with_solutions():
    """Sudoku yield thousands of solution in a second"""
    answers = solve([], inline=SUDOKU, options='--time-limit=1')
    nb_answer = sum(1 for answer in answers)
    assert nb_answer > 1


@pytest.mark.slow
@run_with_clingo_binary_only
def test_time_limit_no_solutions():
    """Queens do not yield any solution in a second
    (at least on not so powerful machines)

    This is tested because clingo do not raise the same error code
    depending of whether a solution was found or not when stopped
    because of the --time-limit option.

    """
    answers = solve([], inline=QUEENS, options='--time-limit=1')
    nb_answer = sum(1 for answer in answers)
    assert nb_answer == 0


@pytest.mark.slow
@run_with_clingo_binary_only
def test_no_time_limit_queens():
    """Queens yield at least one solution when enough time is provided.

    This is tested in order to prove that the time limit option works
    properly in the `test_time_limit_no_solutions` function.

    """
    answers = solve([], inline=QUEENS, nb_model=1)
    assert sum(1 for answer in answers) == 1


QUEENS = """
#const n = 200.
n(1..n).

q(X,Y) :- n(X), n(Y), not not q(X,Y).

        c(r,X; c,Y) :- q(X,Y).
not not c(r,N; c,N) :- n(N).

n(r,X,Y-1,X,Y; c,X-1,Y,X,Y; d1,X-1,Y-1,X,Y;     d2,X-1,Y+1,X,Y      ) :- n(X), n(Y).
c(r,N,0;       c,0,N;       d1,N-1,0; d1,0,N-1; d2,N-1,n+1; d2,0,N+1) :- n(N).

c(C,XX,YY) :-     c(C,X,Y), n(C,X,Y,XX,YY), not q(XX,YY).
           :- not c(C,X,Y), n(C,X,Y,XX,YY),     q(XX,YY).

#show q/2.
"""


SUDOKU = """
val(1..9).
border(1;4;7).

% Only one number per square.
1 { x(X,Y,N) : val(N) } 1 :- val(X) ; val(Y).

% Same numbers does not share row or column…
1 { x(X,Y,N) : val(X) } 1 :- val(N) ; val(Y).
1 { x(X,Y,N) : val(Y) } 1 :- val(N) ; val(X).

% … nor boxes.
1 { x(X,Y,N) : val(X), val(Y), X1<=X, X<=X1+2, Y1<=Y, Y<=Y1+2 } 1 :- val(N) ; border(X1) ; border(Y1).
"""




import pytest
import clyngor
from clyngor.answers import Answers
from clyngor.as_pyasp import Atom, TermSet

@pytest.fixture
def simple_answers():
    return Answers((
        'a(0) b(1)',
        'c(2) d(3)',
        'e(4) f(5)',
        'g(6) h(7)',
        'i j',
    ))

@pytest.fixture
def quotes_answers():
    return Answers((
        'a("c") b("d","e")',
        'a("c") b("d","e")',
        'a("c") b("d","e")',
        'a("c") b("d","e")',
        'a("c") b("d","e")',
        'a("c") b("d","e")',
    ))

@pytest.fixture
def complex_quotes_answers():
    return Answers((
        'a("\\"cou\\"cou\\"") b("\\"&\\"&\\"1234\\"\\"&\\"")',  # Random gibberish
    ))

@pytest.fixture
def multiple_args():
    return Answers((
        'edge(4,"s…lp.") r_e_l(1,2)',
        'edge(4,"s…lp.") r_e_l(1,2)',
        'edge(4,"s…lp.") r_e_l(1,2)',
        'edge(4,"s…lp.") r_e_l(1,2)',
        'edge(4,"s…lp.") r_e_l(1,2)',
    ))

@pytest.fixture
def noarg_answers():
    return Answers((
        '',
        'a',
        'b c',
        'd e f',
    ))

@pytest.fixture
def many_atoms_answers():
    return Answers((
        'a b(4) b(3) a(1) vv(1) vv v(a) v(b)',
    ))

@pytest.fixture
def default_negation_answers():
    return Answers((
        'ok -ko',
        '-occurs(apply(drychem,fire_in(bin)), 2) ok -ko',
    ))

@pytest.fixture
def optimized_answers():
    return Answers((
        ('edge(4,"s…lp.") r_e_l(1,2)', 1, False, 1),
        ('edge(4,"s…lp.") r_e_l(1,2)', 1, False, 2),
        ('edge(4,"s…lp.") r_e_l(1,2)', 2, False, 3),
        ('edge(4,"s…lp.") r_e_l(1,2)', 3, False, 4),
        ('edge(4,"s…lp.") r_e_l(1,2)', 4, False, 5),
    ), with_optimization=True)


def test_parsing_args_when_noargs(noarg_answers):
    answers = noarg_answers
    assert next(answers) == frozenset()
    assert next(answers) == {('a', ())}
    answers = noarg_answers.parse_args
    assert next(answers) == {('b', ()), ('c', ())}
    answers.by_predicate
    assert next(answers) == {'d': frozenset({()}), 'e': frozenset({()}),
                             'f': frozenset({()})}

def test_discard_quotes(quotes_answers):
    answers = quotes_answers
    assert next(answers) == {('a', ('"c"',)), ('b', ('"d"','"e"',))}
    answers = quotes_answers.discard_quotes
    assert next(answers) == {('a', ('c',)), ('b', ('d','e',))}
    answers = quotes_answers.discard_quotes.by_predicate
    assert next(answers) == {'a': frozenset({('c',)}), 'b': frozenset({('d', 'e')})}
    answers = quotes_answers.discard_quotes.by_predicate.first_arg_only
    assert next(answers) == {'a': frozenset({'c'}), 'b': frozenset({'d'})}
    answers = quotes_answers.discard_quotes.by_predicate.first_arg_only.atoms_as_string
    assert next(answers) == {'a("c")', 'b("d")'}
    assert next(answers) == {'a("c")', 'b("d")'}
    assert next(answers, None) is None

def test_discard_quotes_careful_parsing(quotes_answers):
    answers = quotes_answers.careful_parsing
    assert next(answers) == {('a', ('"c"',)), ('b', ('"d"','"e"',))}
    answers = quotes_answers.careful_parsing.discard_quotes
    assert next(answers) == {('a', ('c',)), ('b', ('d','e',))}
    answers = quotes_answers.careful_parsing.discard_quotes.by_predicate
    assert next(answers) == {'a': frozenset({('c',)}), 'b': frozenset({('d', 'e')})}
    answers = quotes_answers.keep_quotes.by_predicate.first_arg_only
    assert next(answers) == {'a': frozenset({'"c"'}), 'b': frozenset({'"d"'})}
    answers = quotes_answers.discard_quotes.by_predicate.first_arg_only
    assert next(answers) == {'a': frozenset({'c'}), 'b': frozenset({'d'})}
    answers = quotes_answers.careful_parsing.discard_quotes.by_predicate.first_arg_only.atoms_as_string
    assert next(answers) == {'a("c")', 'b("d")'}
    assert next(answers, None) is None


def test_discard_quotes_complex(complex_quotes_answers):
    answers = complex_quotes_answers.discard_quotes
    assert next(answers) == {('a', (r'\"cou\"cou\"',)), ('b', (r'\"&\"&\"1234\"\"&\"',))}
    assert next(answers, None) is None

def test_discard_quotes_complex_careful_parsing(complex_quotes_answers):
    answers = complex_quotes_answers.discard_quotes.careful_parsing
    assert next(answers) == {('a', (r'\"cou\"cou\"',)), ('b', (r'\"&\"&\"1234\"\"&\"',))}
    assert next(answers, None) is None


def test_as_pyasp_atom():
    answers = Answers(('a("b","d")',)).as_pyasp
    for atom in next(answers):
        assert atom == Atom(predicate='a', args=('"b"','"d"'))
        print(atom)
    answers = Answers(('a(1,"2.3")',)).as_pyasp
    for atom in next(answers):
        assert atom == Atom(predicate='a', args=(1,'"2.3"'))
        print(atom)

def test_bug_issue_7():
    "inconsistencies in parsing when clingo module is used"
    models = clyngor.solve(inline='a(1). e(X):- a(X).', use_clingo_module=True).sorted.atoms_as_string
    model = next(models)
    assert model == ('a(1)', 'e(1)')
    with pytest.raises(StopIteration):
        next(models)  # only one model
    # assert False, 'everything ok'

def test_as_pyasp_termset_frozenset():
    answers = Answers(('a("b","d")',)).as_pyasp
    assert next(answers) == frozenset((Atom(predicate='a',args=('"b"','"d"')),))

def test_as_pyasp_termset_frozenset_union():
    answers = Answers(('a("b","d")',)).as_pyasp
    answer_a = next(answers)
    expected_a = frozenset((Atom(predicate='a',args=('"b"','"d"')),))
    assert answer_a == expected_a
    assert next(answers, None) is None
    answers = Answers(('b("b","d")',)).as_pyasp
    answer_b = next(answers)
    expected_b = frozenset((Atom(predicate='b',args=('"b"','"d"')),))
    assert answer_b == expected_b
    assert next(answers, None) is None
    assert isinstance(answer_a, TermSet)
    assert isinstance(answer_b, TermSet)
    assert answer_a.union(answer_b) == frozenset((Atom(predicate='a',args=('"b"','"d"')), Atom(predicate='b',args=('"b"','"d"'))))
    # no modification of objects
    assert answer_a == expected_a
    assert answer_b == expected_b

def test_as_pyasp_termset_add():
    answers = Answers(('a("b","d")',)).as_pyasp
    answer_a = next(answers)
    expected_a = frozenset((Atom(predicate='a',args=('"b"','"d"')),))
    assert answer_a == expected_a
    assert next(answers, None) is None
    answer_a.add(Atom(predicate='b',args=('"b"','"d"')))
    expected_a = frozenset((Atom(predicate='a',args=('"b"','"d"')), Atom(predicate='b',args=('"b"','"d"'))))
    assert answer_a == expected_a

def test_as_pyasp_termset_termset():
    answers = Answers(('a("b","d")',)).as_pyasp
    assert next(answers) == TermSet((Atom(predicate='a',args=('"b"','"d"')),))

def test_as_pyasp_termset_string():
    answers = Answers(('a("b","d")',)).as_pyasp
    assert str(next(answers)) == 'a("b","d").'

def test_as_pyasp_termset_termset_by_predicate():
    answers = Answers(('a("b","d")',)).as_pyasp.by_predicate
    assert next(answers) == {'a': TermSet((Atom(predicate='a',args=('"b"','"d"')),))}

def test_discard_quotes_with_as_pyasp_termset():
    answers = Answers(('a("b","d")',)).as_pyasp
    assert next(answers) == TermSet((Atom(predicate='a',args=('"b"','"d"')),))
    answers = Answers(('a("b","d")',)).as_pyasp.discard_quotes
    assert next(answers) == TermSet((Atom(predicate='a',args=('"b"','"d"')),))

def test_discard_quotes_with_as_pyasp_termset_and_careful_parsing():
    answers = Answers(('a("b","d")',)).parse_args.as_pyasp.careful_parsing
    assert next(answers) == TermSet((Atom(predicate='a',args=('"b"','"d"')),))
    answers = Answers(('a("b","d")',)).parse_args.as_pyasp.discard_quotes.careful_parsing
    assert next(answers) == TermSet((Atom(predicate='a',args=('"b"','"d"')),))  # as_pyasp overrides discard_quotes


def test_multiple_tunning_no_arg(noarg_answers):
    answers = noarg_answers.no_arg
    assert next(answers) == frozenset()
    assert next(answers) == {'a'}
    assert next(answers) == {'b', 'c'}
    answers.by_predicate
    assert next(answers) == {'d': frozenset(), 'e': frozenset(), 'f': frozenset()}


def test_by_predicate_and_arity(simple_answers):
    answers = simple_answers.by_predicate
    assert next(answers) == {'a': {(0,)}, 'b': {(1,)}}
    answers.by_arity
    assert next(answers) == {'c': {(2,)}, 'd': {(3,)}, ('c', 1): {(2,)}, ('d', 1): {(3,)}, 'c/1': {(2,)}, 'd/1': {(3,)}}
    assert len(next(answers)) == 6
    assert len(next(answers)) == 6
    assert len(next(answers)) == 6
    assert next(answers, None) is None


def test_multiple_tunning(simple_answers):
    answers = simple_answers
    assert next(answers) == {('a', (0,)), ('b', (1,))}
    answers.by_predicate
    assert next(answers) == {'c': {(2,)}, 'd': {(3,)}}
    answers.first_arg_only
    assert next(answers) == {'e': {4}, 'f': {5}}
    answers.int_not_parsed
    assert next(answers) == {'g': {'6'}, 'h': {'7'}}
    assert next(answers) == {'i': {()}, 'j': {()}}
    assert next(answers, None) is None


def test_tunning_first_arg_only(simple_answers):
    answers = simple_answers
    answers.first_arg_only
    assert next(answers) == {('a', 0), ('b', 1)}
    assert next(answers) == {('c', 2), ('d', 3)}
    assert next(answers) == {('e', 4), ('f', 5)}
    assert next(answers) == {('g', 6), ('h', 7)}
    assert next(answers) == {('i', ()), ('j', ())}
    assert next(answers, None) is None


def test_tunning_atoms_as_string_and_sorted_more_complex(many_atoms_answers):
    answers = many_atoms_answers.atoms_as_string.sorted.first_arg_only
    assert next(answers) == tuple(sorted('a b(4) b(3) a(1) vv(1) vv v(a) v(b)'.split()))
    assert next(answers, None) is None


def test_tunning_atoms_as_string_and_sorted_more_complex_careful_parsing(many_atoms_answers):
    answers = many_atoms_answers.atoms_as_string.sorted.first_arg_only.careful_parsing
    assert next(answers) == tuple(sorted('a b(4) b(3) a(1) vv(1) vv v(a) v(b)'.split()))
    assert next(answers, None) is None


def test_default_negation(default_negation_answers):
    answers = default_negation_answers
    assert next(answers) == {('ok', ()), ('-ko', ())}
    assert next(answers) == {('-occurs', ('apply(drychem,fire_in(bin))', 2)), ('ok', ()), ('-ko', ())}
    assert next(answers, None) is None


def test_first_arg_only():
    answers = Answers(('p(1,2,3)',)).first_arg_only
    assert next(answers) == {('p', 1)}


def test_first_arg_only_with_careful_parsing():
    answers = Answers(('p("1,5","2,456",3)',)).first_arg_only.careful_parsing
    assert next(answers) == {('p', '"1,5"')}


def test_tunning_atoms_as_string_and_sorted(simple_answers):
    answers = simple_answers.atoms_as_string.sorted
    assert next(answers) == ('a(0)', 'b(1)')
    assert next(answers) == ('c(2)', 'd(3)')
    assert next(answers) == ('e(4)', 'f(5)')
    assert next(answers) == ('g(6)', 'h(7)')
    assert next(answers) == ('i', 'j')
    assert next(answers, None) is None


def test_tunning_atoms_as_string(simple_answers):
    answers = simple_answers
    answers.atoms_as_string
    assert next(answers) == {'a(0)', 'b(1)'}
    assert next(answers) == {'c(2)', 'd(3)'}
    answers.careful_parsing
    assert next(answers) == {'e(4)', 'f(5)'}
    assert next(answers) == {'g(6)', 'h(7)'}
    assert next(answers) == {'i', 'j'}
    assert next(answers, None) is None


def test_multiple_args_in_atoms(multiple_args):
    answers = multiple_args
    assert next(answers) == {('edge', (4, '"s…lp."')), ('r_e_l', (1, 2))}
    answers = multiple_args.discard_quotes
    assert next(answers) == {('edge', (4, 's…lp.')), ('r_e_l', (1, 2))}
    answers = answers.atoms_as_string
    assert next(answers) == {'edge(4,"s…lp.")', 'r_e_l(1,2)'}
    answers.careful_parsing
    assert next(answers) == {'edge(4,"s…lp.")', 'r_e_l(1,2)'}
    assert next(answers) == {'edge(4,"s…lp.")', 'r_e_l(1,2)'}
    assert next(answers, None) is None


def test_optimization_access(optimized_answers):
    answers = optimized_answers.with_optimization
    assert next(answers) == ({('edge', (4, '"s…lp."')), ('r_e_l', (1, 2))}, 1)
    answers = optimized_answers.with_optimization.discard_quotes
    assert next(answers) == ({('edge', (4, 's…lp.')), ('r_e_l', (1, 2))}, 1)
    answers = answers.no_arg
    assert next(answers) == ({'edge', 'r_e_l'}, 2)
    answers.atoms_as_string.with_optimality
    assert next(answers) == ({'edge(4,"s…lp.")', 'r_e_l(1,2)'}, 3, False)
    answers.with_answer_number
    assert next(answers) == ({'edge(4,"s…lp.")', 'r_e_l(1,2)'}, 4, False, 5)
    assert next(answers, None) is None

def test_int_not_parsed():
    "Ensures that .int_not_parsed modifier behave correctly"
    asp = 'a(1). #show a/1.'
    for model in next(clyngor.ASP(asp)):
        t = type(model[1][0])
        assert t is int, t
    for model in next(clyngor.ASP(asp).int_not_parsed):
        t = type(model[1][0])
        assert t is str, t

def test_atoms_with_leading_underscore():
    """needed to test leading underscores in naive parsing ;
    next test will fall under the case where careful parsing is automatically activated"""
    get_answers = lambda: Answers(('_a(_b,_c,d) _e',)).parse_args
    # naive parsing
    answers = get_answers()
    answer_a = next(answers)
    assert next(answers, None) is None, "there is more than one model"
    expected_a = frozenset((('_a', ('_b', '_c', 'd')), ('_e', ())))
    assert answer_a == expected_a, (answer_a, type(answer_a))
    # careful parsing
    answers = get_answers().careful_parsing
    answer_a = next(answers)
    assert next(answers, None) is None, "there is more than one model"
    expected_a = frozenset((('_a', ('_b', '_c', 'd')), ('_e', ())))
    assert answer_a == expected_a, answer_a

def test_atoms_with_leading_underscore_with_nested():
    get_answers = lambda: Answers(('_a("b",_b(c,_d)) _c',)).parse_args
    # naive parsing (NB: careful parsing should be automatically activated)
    answers = get_answers()
    answer_a = next(answers)
    assert next(answers, None) is None, "there is more than one model"
    expected_a = frozenset((('_a', ('"b"', ('_b', ('c', '_d')))), ('_c', ())))
    assert answer_a == expected_a, (answer_a, type(answer_a))
    # careful parsing
    answers = get_answers().careful_parsing
    answer_a = next(answers)
    assert next(answers, None) is None, "there is more than one model"
    expected_a = frozenset((('_a', ('"b"', ('_b', ('c', '_d')))), ('_c', ())))
    assert answer_a == expected_a, answer_a

def test_pyasp_and_nested_atoms():
    get_answers = lambda: Answers(('a(b(c,d))',)).parse_args
    # without as_pyasp
    answers = get_answers()
    answer_a = next(answers)
    assert next(answers, None) is None, "there is more than one model"
    expected_a = frozenset((('a', (('b', ('c', 'd')),)),))
    assert answer_a == expected_a, (answer_a, type(answer_a))
    # with as_pyasp
    answers = get_answers().as_pyasp
    answer_a = next(answers)
    assert next(answers, None) is None, "there is more than one model"
    expected_a = frozenset((Atom('a', (Atom('b', ('c', 'd')),)),))
    print('LBTL:', type(answer_a), answer_a)
    assert answer_a == expected_a, (answer_a, type(answer_a))




import tempfile
from .test_api import asp_code  # fixture
from clyngor import solve, ASP, utils


def test_save_in_file(asp_code):
    answers = frozenset(ASP(asp_code))
    with tempfile.NamedTemporaryFile('w', delete=False) as ofd:
        ofd.write('\n'.join(utils.answer_set_to_str(answer) for answer in answers))
        fname = ofd.name

    with open(fname) as ifd:
        read_answers = frozenset(
            frozenset(utils.answer_set_from_str(line))
            for line in ifd
        )

    assert read_answers == answers


def test_file_saving_api(asp_code):
    answers = frozenset(ASP(asp_code))
    fname = utils.save_answers_in_file(answers)
    read_answers = frozenset(utils.load_answers_from_file(fname))
    assert answers == read_answers


def test_file_saving_api_with_dict(asp_code):
    # create the read_answers from the dictionary repr of answer sets
    as_dict_answers = tuple(ASP(asp_code).by_predicate)
    fname = utils.save_answers_in_file(as_dict_answers)
    read_answers = frozenset(utils.load_answers_from_file(fname))

    # must be the same as regular repr of answer sets
    answers = frozenset(ASP(asp_code))
    assert answers == read_answers


def test_file_saving_api_with_as_pyasp(asp_code):
    # create the read_answers from the dictionary repr of answer sets
    as_pyasp_answers = tuple(ASP(asp_code).as_pyasp)
    fname = utils.save_answers_in_file(as_pyasp_answers)
    read_answers = frozenset(utils.load_answers_from_file(fname))

    # must be the same as regular repr of answer sets
    answers = frozenset(ASP(asp_code))
    assert answers == read_answers


def test_remove_quotes_argument():
    example_inputs = ['"\"a\"","b\""', '\\"a\\","b\\"']
    expected_results = ['"a",b"', '\\"a\\","b\\"']

    for found, expected in zip(map(utils.remove_arguments_quotes,example_inputs), expected_results):
        assert found == expected

def test_answer_set_to_str():
    generate_answer_set_as_str = utils.generate_answer_set_as_str
    assert '.'.join(generate_answer_set_as_str((('a', (1, 2)), ('b', ())))) == 'a(1,2).b'
    assert ''.join(generate_answer_set_as_str((('a', (1, 2)), ('b', ())), atom_end='.')) == 'a(1,2).b.'
    assert '1'.join(generate_answer_set_as_str((('a', (1, 2)), ('b', ())), atom_end='2')) == 'a(1,2)21b2'
    assert ' '.join(generate_answer_set_as_str((('', ('v', 2)), ('b', (('', ['v', 3]),))), atom_end='.')) == '(v,2). b((v,3)).'
    assert ' '.join(generate_answer_set_as_str((('', ('', ('', (2,)))),))) == '("",(2,))'
    assert '.'.join(generate_answer_set_as_str((('a', (('', ('1', '2')),)),))) == 'a((1,2))'

def test_answer_set_to_str_with_tuple():
    asp = 'a(b,(2,3,(a,b))).'
    model = next(ASP(asp).parse_args)
    assert ' '.join(utils.generate_answer_set_as_str(model, atom_end='.')) == asp

def test_answer_set_to_str_complex():
    asp = 'a(a(2,3),(2,),b(c((d,),(e,f)))).'
    models = tuple(ASP(asp).parse_args)
    print('CAREFUL:', models)
    answerset = models[0]
    models = tuple(ASP(asp))
    print('NORMAL :', models)
    assert ' '.join(utils.generate_answer_set_as_str(answerset, atom_end='.')) == asp

def test_with_opts():
    ASP = r"""
1 { a; b }.
#minimize { 2,t:a; 2,t:b; 1,u:b; 2,v:b }.
    """
    answer = solve(inline=ASP, options='--opt-mode=optN')
    found = list(utils.opt_models_from_clyngor_answers(answer))
    assert found == [frozenset({('a', ())})]



"""Testing of clyngor.decoder"""

from clyngor import solve, decode


ASP_GEN_CONCEPTS = """  % each model is a concept
% data
rel((a;b),(c;d)).
rel((b;e),(f;g)).
% generation of concepts
extent(X):- rel(X,_) ; rel(X,Y): intent(Y).
intent(Y):- rel(_,Y) ; rel(X,Y): extent(X).
:- not intent(_).  % must be populated
"""
ASP_LIST_CONCEPTS = """  % one model contains all concepts
concept(0).
extent(0,(a;b)). intent(0,(c;d)).
concept(1).
extent(1,(b;e)). intent(1,(f;g)).
concept(2).
extent(2,b). intent(2,(c;d;f;g)).
"""

class ConceptList:
    "Decoder used for tests inside that module using ASP_LIST_CONCEPTS"
    def __init__(self, concept:1, extent:all, intent:all):
        self.id = int(concept[0])
        self.extent = frozenset(arg for nb, arg in extent if nb == self.id)
        self.intent = frozenset(arg for nb, arg in intent if nb == self.id)
    def __eq__(self, othr):
        return all((
            self.id == othr.id,
            self.extent == othr.extent,
            self.intent == othr.intent
        ))
    def __hash__(self):  # ensure equivalence in set
        return hash(tuple(map(hash, (self.id, self.extent, self.intent))))
    def __repr__(self):  # help for debug
        return "<{}: {{{}}} × {{{}}}>".format(self.id, ','.join(sorted(self.extent)), ','.join(sorted(self.intent)))
    @staticmethod
    def expected():
        return {
            ConceptList([0], ([0, 'a'], [0, 'b']), ([0, 'c'], [0, 'd'])),
            ConceptList([1], ([1, 'b'], [1, 'e']), ([1, 'f'], [1, 'g'])),
            ConceptList([2], ([2, 'b'],), ([2, 'c'], [2, 'd'], [2, 'f'], [2, 'g'])),
        }
class ConceptGen(ConceptList):
    "Decoder used for tests inside that module using ASP_GEN_CONCEPTS"
    def __init__(self, extent:all, intent:all):
        self.id = 0  # just here to reuse Concept.List__eq__ (in a real setup, it would be autogenerated)
        self.extent = frozenset(arg for arg, in extent)
        self.intent = frozenset(arg for arg, in intent)
    @staticmethod
    def expected():
        return {
            ConceptGen((['a'], ['b']), (['c'], ['d'])),
            ConceptGen((['b'], ['e']), (['f'], ['g'])),
            ConceptGen((['b'],), (['c'], ['d'], ['f'], ['g'])),
        }

def test_Concept_classes():
    "just to ensure the important property of the object"
    assert ConceptList([0], ([0, 'a'],), ([0, 'c'],)) == ConceptList([0], ([0, 'a'],), ([0, 'c'],))
    assert ConceptGen((['a'],), (['c'],)) == ConceptGen((['a'],), (['c'],))

def test_decoder_list_concepts_per_decode():
    "All Concept in one model, first method: directly get decoded concepts"
    objects = set(decode(inline=ASP_LIST_CONCEPTS, decoders=[ConceptList]))
    assert objects == ConceptList.expected()

# def test_decoder_list_concepts_per_solve():
    # "All Concept in one model, second method: inside the solving scheme"
    # models = solve(inline=ASP_LIST_CONCEPTS, decoders=[ConceptList])
    # assert set(models.all_objects()) == ConceptList.expected()

def test_decoder_gen_concepts_per_decode():
    "One Concept per model, first method: directly get decoded concepts"
    objects = set(decode(inline=ASP_GEN_CONCEPTS, decoders=[ConceptGen]))
    assert objects == ConceptGen.expected()

# def test_decoder_gen_concepts_per_solve():
    # "One Concept per model, first method: inside the solving scheme"
    # models = solve(inline=ASP_GEN_CONCEPTS, decoders=[ConceptGen])
    # assert set(models.all_objects()) == ConceptGen.expected()



"""Some definitions used for testing.

"""
import pytest
import clyngor
from functools import wraps


def run_with_clingo_binary_only(func):
    """Decorator deactivating clingo module handling before running
    the test function, then reactivating it if available.

    """
    @wraps(func)
    def wrapped(*args, **kwargs):
        if clyngor.clingo_module_actived():
            clyngor.deactivate_clingo_module()
            ret = func(*args, **kwargs)
            clyngor.use_clingo_module()
        else:  # clingo module not here, so there is nothing to do
            ret = func(*args, **kwargs)
        return ret
    return wrapped


def run_with_clingo_module_only(func):
    """Decorator activating clingo module handling before running
    the test function, then deactivating it if available.

    """
    @wraps(func)
    def wrapped(*args, **kwargs):
        was_using_module = clyngor.clingo_module_actived()
        clyngor.use_clingo_module()
        ret = func(*args, **kwargs)
        if not was_using_module:
            clyngor.use_clingo_binary()
        return ret
    return onlyif_clingo_module(wrapped)


def skipif_clingo_without_python(func):
    return pytest.mark.skipif(
        not clyngor.have_python_support(py3=True),
        reason="Require clingo with python3 support"
    )(func)


def skipif_no_clingo_module(func):
    return pytest.mark.skipif(
        not clyngor.have_clingo_module(),
        reason="Require official clingo module to be available"
    )(func)


# clearer names and oppositions
onlyif_clingo_module = skipif_no_clingo_module
onlyif_python_support = skipif_clingo_without_python

def onlyif_no_python_support(func):
    return pytest.mark.skipif(
        clyngor.have_python_support(py3=True),
        reason="Requires clingo not to support python"
    )(func)

def onlyif_no_clingo_module(func):
    return pytest.mark.skipif(
        clyngor.have_clingo_module(),
        reason="Require official clingo module to NOT be available"
    )(func)





import pytest
import clyngor
from .definitions import onlyif_clingo_module, onlyif_no_clingo_module, onlyif_python_support, onlyif_no_python_support, run_with_clingo_module_only, run_with_clingo_binary_only

@run_with_clingo_module_only
@onlyif_clingo_module
def test_basic_example():
    "Example taken from clingo module documentation: https://potassco.org/clingo/python-api/5.5/clingo/index.html"
    from clingo.symbol import Number
    from clingo.control import Control
    class Context:
        def inc(self, x):
            return Number(x.number + 1)
        def seq(self, x, y):
            return [x, y]

    def on_model(m):
        print(repr(m), m, dir(m))
        assert False, "Clingo module seems to have python support. Some test on the output model are to be done (just be sure there is a `a` atom). Model: " + repr(m)
        assert clyngor.have_python_support()

    ctl = Control()
    try:
        ctl.add("base", [], "a. #script(python) #end.")
    except RuntimeError as err:  # case where python support is not implemented
        assert err.args == ('<block>:1:4-25: error: python support not available\n',)
        assert not clyngor.have_python_support()
    else:  # python support available
        ctl.ground([("base", [])], context=Context())
        ctl.solve(on_model=on_model)


@run_with_clingo_module_only
@onlyif_clingo_module
@onlyif_python_support
def test_clingo_module_detection_and_state_MP():
    assert clyngor.utils.try_python_availability_in_clingo()
    assert clyngor.utils.try_python_availability_in_clingo_module()

@run_with_clingo_module_only
@onlyif_clingo_module
@onlyif_no_python_support
def test_clingo_module_detection_and_state_MnP():
    assert not clyngor.utils.try_python_availability_in_clingo()
    assert not clyngor.utils.try_python_availability_in_clingo_module()

@run_with_clingo_binary_only
@onlyif_no_clingo_module
@onlyif_python_support
def test_clingo_module_detection_and_state_BP():
    assert     clyngor.utils.try_python_availability_in_clingo()
    assert     clyngor.utils.try_python_availability_in_clingo_binary()

@run_with_clingo_binary_only
@onlyif_no_clingo_module
@onlyif_no_python_support
def test_clingo_module_detection_and_state_BnP():
    assert not clyngor.utils.try_python_availability_in_clingo()
    assert not clyngor.utils.try_python_availability_in_clingo_binary()
    with pytest.raises(ImportError):
        clyngor.utils.try_python_availability_in_clingo_module()




"""Decorators and functions for python embedded ASP source code.


"""

import math
import inspect
import clyngor
from functools import wraps


def _converted_types(ignore_bad_type:bool=True):
    """Decorator for functions accessed by ASP, using type annotations to
    convert input types in their expected type.

    Also, if the decorated function is a generator, it will be transformed into a
    list in order to match clingo's API.
    Non-conformant types are sent to the function without modification.

    ignore_bad_type -- Non-conformant types are ignored.

    """
    if not clyngor.have_python_support() or not clyngor.clingo_module_available:
        return clyngor.utils.null_decorator
    from clingo import symbol
    def clingo_to_python(val:symbol, annot:type) -> object:
        if annot is None or not type(val).__name__ == 'Symbol':
            return val
        elif annot is str and val.type == val.type.String:
            return val.string
        elif annot is int and val.type == val.type.Number:
            return val.number
        elif annot is int and val.type == val.type.Infimum:
            return -math.inf
        elif annot is int and val.type == val.type.Supremum:
            return math.inf
        elif ignore_bad_type:
            raise TypeError(f"Bad type hint match from ASP: received {type(val)}, annotation is {annot}")
        else:
            return val

    def output_to_clingo(val: object) -> symbol:
        if isinstance(val, symbol.Symbol):
            return val
        elif isinstance(val, str):
            return symbol.String(val)
        elif isinstance(val, int):
            return symbol.Number(val)
        elif val == math.inf:
            return symbol.Infimum()
        elif val == -math.inf:
            return symbol.Supremum()
        elif isinstance(val, list):
            return list(map(output_to_clingo, val))
        elif isinstance(val, tuple):
            return symbol.Function('', tuple(map(output_to_clingo, val)))
        elif type(val).__name__ == 'generator':
            return output_to_clingo(list(val))
        elif ignore_bad_type:
            raise TypeError(f"Function outputed value {repr(val)} of type {type(val)}, which is not handled.")
        else:
            return val

    def convert_args(argspec, args, kwargs) -> (str, object):
        for name, arg in zip(argspec.args, args):
            annot = argspec.annotations.get(name)
            yield name, clingo_to_python(arg, annot)
        for name, arg in kwargs.items():
            annot = argspec.annotations.get(name)
            yield name, clingo_to_python(arg, annot)

    def decorator(func):
        argspec = inspect.getfullargspec(func)
        isgenerator = inspect.isgeneratorfunction(func)

        @wraps(func)
        def decorated(*args, **kwargs):
            try:
                ret = func(**dict(convert_args(argspec, args, kwargs)))
            except TypeError:
                return []  # nothing to be done
            return_type = argspec.annotations.get('return')
            if callable(return_type):
                ret = return_type(ret)
            elif isgenerator:
                ret = list(ret)  # each returned object is an independant value.
            ret = output_to_clingo(ret)
            return ret

        return decorated
    return decorator

converted_types = _converted_types(True)
converted_types_or_symbols = _converted_types(False)



import pytest

from clyngor import parsing, answer_set_to_str
from clyngor.parsing import Parser


def test_careful_parsing_required():
    yeses = (
        'a(b(1)).',
        'a("hello ; a(b)").',
        'a("hello,b").',
        'a((1,2,3),4).',
    )
    nos = (
        'a(1,2,3).',
        'a(1,"hello",3).',
        'a("hello ",3). b("").',
    )
    for yes in yeses:
        print('YES:', yes)
        assert     parsing.careful_parsing_required(yes)
    for no in nos:
        print('NO :', no)
        assert not parsing.careful_parsing_required(no)


def test_simple_case():
    parsed = Parser().parse_clasp_output(OUTCLASP_SIMPLE.splitlines(),
                                         yield_stats=True, yield_info=True)
    models = []
    for type, payload in parsed:
        assert type in ('statistics', 'info', 'answer', 'answer_number')
        if type == 'statistics':
            stats = payload
        elif type == 'answer':
            models.append(payload)
        elif type == 'answer_number':
            pass
        else:
            assert type == 'info'
            info = payload
    assert len(models) == 2
    assert info == ('clasp version 3.2.0', 'Reading from stdin', 'Solving...', 'SATISFIABLE')
    assert stats == {'Models': '2', 'Calls': '1', 'CPU Time': '0.000s',
                     'Time': '0.001s (Solving: 0.00s 1st Model: 0.00s Unsat: 0.00s)'}

def test_parse_default_negation():
    string = '-r(1) -r(-2)'
    expected = {
        ('-r', (1,)), ('-r', (-2,)),
    }
    assert Parser().parse_terms(string) == expected


def test_parse_termset_default():
    string = r'a(b,10) c(d("a",d_d),"v,\"v\"",c) d(-2,0)'
    expected = {
        ('a', ('b', 10)),
        ('c', ('d("a",d_d)', r'"v,\"v\""', 'c')),
        ('d', (-2, 0)),
    }
    found = Parser().parse_terms(string)
    print(frozenset(expected))
    print(found)
    assert found == expected

def test_parse_tuple():
    string = 'a(b,(2,3,(a,b)))'
    expected = {
        ('a', ('b', ('', (2, 3, ('', ('a', 'b')))))),
    }
    assert Parser().parse_terms(string) == expected

def test_parse_empty_string_arg():
    string = 'b("")'
    expected = {
        ('b', ('""',)),
    }
    assert Parser().parse_terms(string) == expected
    assert sorted(answer_set_to_str(expected).split()) == sorted(string.split())

def test_parse_string_with_space():
    string = 'a("hello ",3) b("")'
    expected = {
        ('a', ('"hello "', 3)),
        ('b', ('""',)),
    }
    assert Parser().parse_terms(string) == expected
    str_reprs = ['b("") a("hello ",3)', 'a("hello ",3) b("")']
    assert answer_set_to_str(expected) in str_reprs

def test_parse_args_without_predicat():
    string = 'a((b,10)) c("",("",""))'
    expected = {
        ('a', (('', ('b', 10)),)),
        ('c', ('""', ('', ('""', '""')))),
    }
    assert Parser().parse_terms(string) == expected
    assert sorted(answer_set_to_str(expected).split()) == sorted(string.split())

def test_parse_termset():
    string = 'a(b,10) c(d("a",d_d),"v,v",c) d(-2,0)'
    expectation = {
        (False, True): {
            ('a', ('b', 10)),
            ('c', ('d("a",d_d)', '"v,v"', 'c')),
            ('d', (-2, 0)),
        },
        (False, False): {
            ('a', ('b', 10)),
            ('c', (('d', ('"a"', 'd_d')), '"v,v"', 'c')),
            ('d', (-2, 0)),
        },
        (True, True): {
            'a(b,10)',
            'c(d("a",d_d),"v,v",c)',
            'd(-2,0)',
        },
        # False True is expected to raise an Error (see dedicated function)
    }
    for parser_mode, expected in expectation.items():
        print(*parser_mode)
        assert    Parser(*parser_mode).parse_terms(string) == frozenset(expected)

def test_parse_termset_impossible():
    with pytest.raises(ValueError) as e_info:
        Parser(True, False)


def test_string():
    """Show that string with comma in it is handled correctly"""
    parsed = Parser().parse_clasp_output(OUTCLASP_STRING.splitlines())
    type, answer_number = next(parsed)
    assert type == 'answer_number'
    type, model = next(parsed)
    assert next(parsed, None) is None, "there is only one model"
    assert type == 'answer', "the model is an answer"
    assert len(model) == 1, "only 1 atom in it"
    pred, args = next(iter(model))
    assert pred == 'atom'
    assert args == ('","',)


def test_single_litteral():
    """Show that string with comma in it is handled correctly"""
    parsed = Parser().parse_clasp_output(OUTCLASP_SINGLE_LITERAL.splitlines())
    type, answer_number = next(parsed)
    assert type == 'answer_number'
    type, model = next(parsed)
    assert next(parsed, None) is None, "there is only one model"
    assert type == 'answer', "the model is an answer"
    assert len(model) == 2, "only 2 atom in it"
    assert model == {3, '"hello !"'}

def test_complex_atoms():
    parsed = Parser().parse_clasp_output(OUTCLASP_COMPLEX_ATOMS.splitlines())
    type, answer_number = next(parsed)
    assert type == 'answer_number'
    type, model = next(parsed)
    assert next(parsed, None) is None, "there is only one model"
    assert type == 'answer', "the model is an answer"
    assert len(model) == 3, "only 3 atoms in it"
    atoms = {atom[0]: atom for atom in model}
    assert set(atoms.keys()) == {'a', 'b', 'c'}


def test_optimization():
    parsed = Parser().parse_clasp_output(OUTCLASP_OPTIMIZATION.splitlines(), yield_stats=True)
    expected_stats = {
        'CPU Time': '0.130s',
        'Calls': '1',
        'Models': '1',
        'Optimization': '190',
        'Optimum': 'yes',
        'Threads': '4        (Winner: 2)',
        'Time': '0.051s (Solving: 0.03s 1st Model: 0.00s Unsat: 0.03s)'
    }
    for type, model in parsed:
        if type == 'statistics':
            assert model == expected_stats


def test_time_limit():
    parsed = Parser().parse_clasp_output(OUTCLASP_TIME_LIMIT.splitlines(),
                                         yield_stats=True,
                                         yield_info=True)
    expected_stats = {
        'TIME LIMIT': '1',
        'Models': '3+',
        'Optimum': 'unknown',
        'Optimization': '597301577',
        'Calls': '1',
        'Time': '4.000s (Solving: 2.82s 1st Model: 0.01s Unsat: 0.00s)',
        'CPU Time': '3.980s',
    }
    expected_info = iter((
        tuple(OUTCLASP_TIME_LIMIT.splitlines()[:3] + ['SATISFIABLE']),
    ))
    expected_answer = iter((
        (('a', ()),), (('b', ()),), (('c', ()),)
    ))
    expected_optimization = iter((597337713, 597301761, 597301577))
    for type, model in parsed:
        if type == 'statistics':
            assert model == expected_stats
        elif type == 'info':
            assert model == next(expected_info)
        elif type == 'answer':
            assert model == frozenset(next(expected_answer))
        elif type == 'answer_number':
            assert isinstance(model, int) and model > 0, model
        elif type == 'optimization':
            assert model == (next(expected_optimization),)
        elif type == 'progression':
            assert False, 'no progression'
        else:  # impossible
            assert False
    # ensure that all data has been found
    assert next(expected_info, None) is None
    assert next(expected_answer, None) is None
    assert next(expected_optimization, None) is None


def test_multiple_opt_values():
    parsed = Parser().parse_clasp_output(CLINGO_OUTPUT_MULTIPLE_OPT.splitlines())
    expected_answer = iter((
        (('a', ()),),
        (('b', ()),),
        (('c', ()),),
    ))
    expected_optimization = iter(((-10, 6), (-11, 6), (-13, 4)))
    for type, model in parsed:
        if type == 'statistics':
            assert False, 'no statistics'
        elif type == 'info':
            assert False, 'no info'
        elif type == 'answer':
            assert model == frozenset(next(expected_answer))
        elif type == 'answer_number':
            assert isinstance(model, int) and model > 0, model
        elif type == 'optimization':
            assert model == next(expected_optimization)
        elif type == 'optimum found':
            assert model == True
        else:  # impossible
            assert False
    # ensure that all data has been found
    assert next(expected_answer, None) is None
    assert next(expected_optimization, None) is None


def test_unsat():
    parsed = Parser().parse_clasp_output(OUTCLASP_UNSATISFIABLE.splitlines())
    current = next(parsed, None)
    assert current is not None
    l_type, model = current
    assert l_type == 'unsat' and model is True
    assert next(parsed, None) is None


def test_unsat_stats():
    parsed = Parser().parse_clasp_output(OUTCLASP_UNSATISFIABLE.splitlines(), yield_stats=True)
    stats = {
        'Models': '0',
        'Calls': '1',
        'Time': '0.001s (Solving: 0.00s 1st Model: 0.00s Unsat: 0.00s)',
        'CPU Time': '0.000s'
    }

    for type, model in parsed:
        assert type in ('unsat', 'statistics')
        if type == 'unsat':
            assert model is True
        elif type == 'statistics':
            assert model == stats


def test_multithread_with_progression():
    parsed = Parser().parse_clasp_output(CLINGO_OUTPUT_MULTITHREADING_WITH_PROGRESS.splitlines(),
                                         yield_prgs=True)
    expected_answer = iter((
        (('a', ()),),
        (('b', ()),),
    ))
    expected_progress = iter((
        '[   2;3299] (Error: 1648.5)',
    ))
    expected_optimization = iter(((3299,), (3296,)))
    for type, model in parsed:
        if type == 'statistics':
            assert False, 'no statistics'
        elif type == 'info':
            assert False, 'no info'
        elif type == 'answer':
            assert model == frozenset(next(expected_answer))
        elif type == 'answer_number':
            assert isinstance(model, int) and model > 0, model
        elif type == 'optimization':
            assert model == next(expected_optimization)
        elif type == 'progression':
            assert model == next(expected_progress)
        else:  # impossible
            assert False
    # ensure that all data has been found
    assert next(expected_answer, None) is None
    assert next(expected_progress, None) is None
    assert next(expected_optimization, None) is None


OUTCLASP_TIME_LIMIT = """clingo version 4.5.4
Reading from search.lp ...
Solving...
Answer: 1
a
Optimization: 597337713
Answer: 2
b
Optimization: 597301761
Answer: 3
c
Optimization: 597301577
SATISFIABLE

TIME LIMIT   : 1
Models       : 3+
  Optimum    : unknown
Optimization : 597301577
Calls        : 1
Time         : 4.000s (Solving: 2.82s 1st Model: 0.01s Unsat: 0.00s)
CPU Time     : 3.980s
"""

OUTCLASP_COMPLEX_ATOMS = """clasp version 3.2.0
Reading from stdin
Solving...
Answer: 1
a(b(1,"a")) b(42,12) c("42,12,a(23)",c)
SATISFIABLE

Models       : 1
Calls        : 1
Time         : 0.001s (Solving: 0.00s 1st Model: 0.00s Unsat: 0.00s)
CPU Time     : 0.000s
"""

OUTCLASP_SINGLE_LITERAL = """clasp version 3.2.0
Reading from test.lp
Solving...
Answer: 1
3 "hello !"
SATISFIABLE

Models       : 1
Calls        : 1
Time         : 0.001s (Solving: 0.00s 1st Model: 0.00s Unsat: 0.00s)
CPU Time     : 0.001s
"""

OUTCLASP_STRING = """clasp version 3.2.0
Reading from test.lp
Solving...
Answer: 1
atom(",")
SATISFIABLE

Models       : 1
Calls        : 1
Time         : 0.001s (Solving: 0.00s 1st Model: 0.00s Unsat: 0.00s)
CPU Time     : 0.001s
"""

OUTCLASP_SIMPLE = """clasp version 3.2.0
Reading from stdin
Solving...
Answer: 1
a
Answer: 2
b
SATISFIABLE

Models       : 2
Calls        : 1
Time         : 0.001s (Solving: 0.00s 1st Model: 0.00s Unsat: 0.00s)
CPU Time     : 0.000s
"""

OUTCLASP_OPTIMIZATION = """clasp version 3.2.0
Reading from l.lp
Solving...
Answer: 1
set(2,9) set(2,10) score(20)
Optimization: 190
OPTIMUM FOUND

Models       : 1
  Optimum    : yes
Optimization : 190
Calls        : 1
Time         : 0.051s (Solving: 0.03s 1st Model: 0.00s Unsat: 0.03s)
CPU Time     : 0.130s
Threads      : 4        (Winner: 2)
"""

OUTCLASP_UNSATISFIABLE = """clasp version 3.2.0
Reading from l.lp
Solving...
UNSATISFIABLE

Models       : 0
Calls        : 1
Time         : 0.001s (Solving: 0.00s 1st Model: 0.00s Unsat: 0.00s)
CPU Time     : 0.000s
"""


CLINGO_OUTPUT_MULTIPLE_OPT = """
clingo version 5.2.2
Reading from ...ergrasp/ASPsources/findbestbiclique.lp ...
Solving...
Answer: 1
a
Optimization: -10 6
Answer: 2
b
Optimization: -11 6
Answer: 3
c
Optimization: -13 4
OPTIMUM FOUND

Models       : 3
  Optimum    : yes
Optimization : -13 4
Calls        : 1
Time         : 0.012s (Solving: 0.00s 1st Model: 0.00s Unsat: 0.00s)
CPU Time     : 0.011s
"""


CLINGO_OUTPUT_MULTITHREADING_WITH_PROGRESS = """
clingo version 5.2.2
Reading from ...rasp/powergrasp/asp/search-biclique.lp ...
/home/lbourneu/packages/powergrasp/powergrasp/asp/process-motif.lp:39:31-53: info: atom does not occur in any rule head:
  include_block(K,T,L,U)

Solving...
Answer: 1
a
Optimization: 3299
Progression : [   2;3299] (Error: 1648.5)
Answer: 2
b
Optimization: 3296
"""



"""Testing of the decoupled grounding/solving API"""

import pytest
from clyngor import grounded_program, solve_from_grounded, solve as solve_standard, opt_models_from_clyngor_answers
from .definitions import run_with_clingo_binary_only


@run_with_clingo_binary_only
def test_simple():
    prg = '1{p(a;b;c)}1.'
    grounded = grounded_program(inline=prg)
    found = frozenset(solve_from_grounded(grounded))  # clingo module cannot be used here
    expected = frozenset(solve_standard(inline=prg))
    assert found == expected


@run_with_clingo_binary_only
def test_with_opts():
    ASP = r"""
q(1..10).
1{p(X): q(X)}3.
nb(odd,N):- N={p(X): X\2=0}.
nb(even,N):- N={p(X): X\2=1}.
nb(sum,N):- N=#sum{X:p(X)}.
#minimize{N,2:nb(odd,N)}.
#maximize{N,1:nb(even,N)}.
    """
    grounded = grounded_program(inline=ASP)
    found = frozenset(opt_models_from_clyngor_answers(solve_from_grounded(grounded).by_predicate))
    expected = frozenset(opt_models_from_clyngor_answers(solve_standard(inline=ASP).by_predicate))
    assert found == expected




import pytest
from .test_api import asp_code  # fixture
import clyngor
from clyngor import solve
from .definitions import run_with_clingo_binary_only, run_with_clingo_module_only


@pytest.fixture
def asp_code_with_constants():
    return """
    #const a=1.
    p(a).
    q(X):- p(X).
    """


@run_with_clingo_binary_only
def test_without_constants(asp_code_with_constants):
    answers = tuple(solve([], inline=asp_code_with_constants).by_predicate)
    assert len(answers) == 1
    answer = answers[0]['q']
    assert len(answer) == 1
    assert next(iter(answer)) == (1,)


@run_with_clingo_binary_only
def test_constants(asp_code_with_constants):
    answers = tuple(solve([], inline=asp_code_with_constants,
                          constants={'a': 2}, print_command=True).by_predicate)
    assert len(answers) == 1
    answer = answers[0]['q']
    assert len(answer) == 1
    assert next(iter(answer)) == (2,)


LITERALS_ARE_SHOWN = 'a.  link(a).  #show link/1.  #show 3. #show "hello !".'

@run_with_clingo_module_only
def test_literal_outputs_by_show():
    with pytest.raises(RuntimeError) as excinfo:
        # If this fails, there is some chance the clingo module
        #  now accepts litterals in output.
        answers = tuple(solve(inline=LITERALS_ARE_SHOWN).by_predicate)

@run_with_clingo_binary_only
def test_literal_outputs_by_show_working():
    answers = tuple(solve(inline=LITERALS_ARE_SHOWN).by_predicate)
    print(answers)
    assert len(answers) == 1
    answer = answers[0]
    print(answers)
    assert len(answer) == 3
    assert set(answer) == {'link', '"hello !"', 3}
    assert answer == {'link': {('a',)}, '"hello !"': {()}, 3: {()}}


# TODO: test solving.command
